# 11 "tools/coqdoc/cpretty.mll"
 
  open Printf
  open Lexing

  (* A list function we need *)
  let rec take n ls =
    if n = 0 then [] else
      match ls with
        | [] -> []
        | (l :: ls) -> l :: (take (n-1) ls)

  (* count the number of spaces at the beginning of a string *)
  let count_spaces s =
    let n = String.length s in
    let rec count c i =
      if i == n then c,i else match s.[i] with
	| '\t' -> count (c + (8 - (c mod 8))) (i + 1)
	| ' ' -> count (c + 1) (i + 1)
	| _ -> c,i
    in
      count 0 0

  let remove_newline s =
    let n = String.length s in
    let rec count i = if i == n || s.[i] <> '\n' then i else count (i + 1) in
    let i = count 0 in
    i, String.sub s i (n - i)

  let count_dashes s =
    let c = ref 0 in
    for i = 0 to String.length s - 1 do if s.[i] = '-' then incr c done;
    !c

  let cut_head_tail_spaces s =
    let n = String.length s in
    let rec look_up i = if i == n || s.[i] <> ' ' then i else look_up (i+1) in
    let rec look_dn i = if i == -1 || s.[i] <> ' ' then i else look_dn (i-1) in
    let l = look_up 0 in
    let r = look_dn (n-1) in
    if l <= r then String.sub s l (r-l+1) else s

  let sec_title s =
    let rec count lev i =
      if s.[i] = '*' then
	count (succ lev) (succ i)
      else
	let t = String.sub s i (String.length s - i) in
	lev, cut_head_tail_spaces t
    in
    count 0 (String.index s '*')

  let strip_eol s =
    let eol = s.[String.length s - 1] = '\n' in
    (eol, if eol then String.sub s 1 (String.length s - 1) else s)


  let formatted = ref false
  let brackets = ref 0
  let comment_level = ref 0
  let in_proof = ref None

  let in_env start stop = 
    let r = ref false in
    let start_env () = r := true; start () in
    let stop_env () = if !r then stop (); r := false in
      (fun x -> !r), start_env, stop_env
      
  let _, start_emph, stop_emph = in_env Output.start_emph Output.stop_emph
  let in_quote, start_quote, stop_quote = in_env Output.start_quote Output.stop_quote

  let url_buffer = Buffer.create 40
  let url_name_buffer = Buffer.create 40

  let backtrack lexbuf = lexbuf.lex_curr_pos <- lexbuf.lex_start_pos;
    lexbuf.lex_curr_p <- lexbuf.lex_start_p

  let backtrack_past_newline lexbuf =
    let buf = lexeme lexbuf in
    let splits = Str.bounded_split_delim (Str.regexp "['\n']") buf 2 in
      match splits with
        | [] -> ()
        | (_ :: []) -> ()
        | (s1 :: rest :: _) ->
            let length_skip = 1 + String.length s1 in
              lexbuf.lex_curr_pos <- lexbuf.lex_start_pos + length_skip

  (* saving/restoring the PP state *)

  type state = {
    st_gallina : bool;
    st_light : bool
  }

  let state_stack = Stack.create ()

  let save_state () =
    Stack.push { st_gallina = !Cdglobals.gallina; st_light = !Cdglobals.light } state_stack

  let restore_state () =
    let s = Stack.pop state_stack in
    Cdglobals.gallina := s.st_gallina;
    Cdglobals.light := s.st_light

  let begin_show () = save_state (); Cdglobals.gallina := false; Cdglobals.light := false
  let end_show () = restore_state ()

  (* Reset the globals *)

  let reset () =
    formatted := false;
    brackets := 0;
    comment_level := 0

  (* erasing of Section/End *)

  let section_re = Str.regexp "[ \t]*Section"
  let end_re = Str.regexp "[ \t]*End"
  let is_section s = Str.string_match section_re s 0
  let is_end s = Str.string_match end_re s 0

  let sections_to_close = ref 0

  let section_or_end s =
    if is_section s then begin
      incr sections_to_close; true
    end else if is_end s then begin
      if !sections_to_close > 0 then begin
	decr sections_to_close; true
      end else
	false
    end else
      true

  (* for item lists *)
  type list_compare =
      | Before
      | StartLevel of int
      | InLevel of int * bool

  (* Before : we're before any levels
     StartLevel : at the same column as the dash in a level
     InLevel : after the dash of this level, but before any deeper dashes.
               bool is true if this is the last level *)
  let find_level levels cur_indent =
    match levels with
    | [] -> Before
    | (l::ls) ->
        if cur_indent < l then Before
        else
          (* cur_indent will never be less than the head of the list *)
          let rec findind ls n =
            match ls with
            | [] -> InLevel (n,true)
            | (l :: []) -> if cur_indent = l then StartLevel n
                           else InLevel (n,true)
            | (l1 :: l2 :: ls) ->
                if cur_indent = l1 then StartLevel n
                else if cur_indent < l2 then InLevel (n,false)
                     else findind (l2 :: ls) (n+1)
          in
            findind (l::ls) 1

  type is_start_list =
    | Rule
    | List of int
    | Neither

  let check_start_list str =
    let n_dashes = count_dashes str in
    let (n_spaces,_) = count_spaces str in
      if n_dashes >= 4 && not !Cdglobals.plain_comments then
        Rule
      else
        if n_dashes = 1 && not !Cdglobals.plain_comments then
          List n_spaces
        else
          Neither

   (* examine a string for subtitleness *)
  let subtitle m s =
    match Str.split_delim (Str.regexp ":") s with
    | [] -> false
    | (name::_) ->
        if (cut_head_tail_spaces name) = m then
          true
        else
          false


  (* tokens pretty-print *)

  let token_buffer = Buffer.create 1024

  let token_re =
    Str.regexp "[ \t]*(\\*\\*[ \t]+printing[ \t]+\\([^ \t]+\\)"
  let printing_token_re =
    Str.regexp
      "[ \t]*\\(\\(%\\([^%]*\\)%\\)\\|\\(\\$[^$]*\\$\\)\\)?[ \t]*\\(#\\(\\(&#\\|[^#]\\)*\\)#\\)?"

  let add_printing_token toks pps =
    try
      if Str.string_match token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	if Str.string_match printing_token_re pps 0 then
	  let pp =
	    (try Some (Str.matched_group 3 pps) with _ ->
             try Some (Str.matched_group 4 pps) with _ -> None),
	    (try Some (Str.matched_group 6 pps) with _ -> None)
	  in
	  Output.add_printing_token tok pp
    with _ ->
      ()

  let remove_token_re =
    Str.regexp
      "[ \t]*(\\*\\*[ \t]+remove[ \t]+printing[ \t]+\\([^ \t]+\\)[ \t]*\\*)"

  let remove_printing_token toks =
    try
      if Str.string_match remove_token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	Output.remove_printing_token tok
    with _ ->
      ()

  let output_indented_keyword s lexbuf =
    let nbsp,isp = count_spaces s in
    Output.indentation nbsp;
    let s = String.sub s isp (String.length s - isp) in
    Output.keyword s (lexeme_start lexbuf + isp)

  let only_gallina () =
    !Cdglobals.gallina && !in_proof <> None

  let parse_comments () =
    !Cdglobals.parse_comments && not (only_gallina ())


# 241 "tools/coqdoc/cpretty.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\236\255\237\255\000\000\000\000\001\000\000\000\000\000\
    \002\000\002\000\005\000\004\000\007\000\022\000\014\000\027\000\
    \024\000\017\000\004\000\031\000\036\000\037\000\052\000\000\000\
    \001\000\002\000\141\000\021\000\010\000\022\000\057\000\041\000\
    \058\000\064\000\065\000\024\000\075\000\083\000\065\000\073\000\
    \063\000\080\000\008\000\128\000\056\000\128\000\143\000\001\000\
    \005\000\221\000\223\000\036\000\064\000\227\000\007\000\254\255\
    \226\000\074\000\074\000\232\000\084\000\084\000\079\000\091\000\
    \089\000\098\000\009\000\012\000\013\000\014\000\015\000\093\000\
    \092\000\087\000\114\000\016\000\239\000\102\000\128\000\132\000\
    \129\000\141\000\139\000\147\000\017\000\018\000\019\000\021\000\
    \022\000\253\000\011\000\240\255\149\000\152\000\150\000\024\000\
    \008\001\159\000\161\000\167\000\171\000\043\001\014\000\025\000\
    \252\255\009\000\165\000\159\000\044\001\017\000\045\001\251\255\
    \019\000\179\000\027\000\047\001\176\000\170\000\163\000\051\001\
    \033\000\057\001\250\255\027\000\216\000\245\255\173\000\218\000\
    \001\001\170\000\231\000\208\000\214\000\209\000\225\000\235\000\
    \233\000\232\000\235\000\087\001\253\255\246\000\247\000\004\001\
    \245\000\239\000\248\255\241\000\250\000\252\000\003\001\255\000\
    \253\000\014\001\255\000\018\001\002\001\015\001\003\001\021\001\
    \023\001\024\001\027\001\011\001\011\001\015\001\029\001\011\001\
    \016\001\034\001\029\001\024\001\026\001\022\001\023\001\038\001\
    \249\255\025\001\030\001\032\001\038\001\045\001\048\001\038\001\
    \035\001\049\001\042\001\034\001\050\001\038\001\044\001\057\001\
    \058\001\056\001\061\001\051\001\060\001\154\001\155\001\064\001\
    \052\001\067\001\061\001\058\001\066\001\052\001\072\001\063\001\
    \066\001\064\001\069\001\064\001\076\001\083\001\078\001\068\001\
    \176\001\179\001\075\001\087\001\074\001\083\001\092\001\084\001\
    \079\001\081\001\081\001\101\001\099\001\104\001\094\001\194\001\
    \197\001\088\001\092\001\090\001\104\001\096\001\111\001\101\001\
    \103\001\101\001\100\001\107\001\102\001\210\001\211\001\110\001\
    \115\001\118\001\118\001\124\001\134\001\116\001\125\001\144\001\
    \132\001\137\001\133\001\138\001\136\001\146\001\138\001\154\001\
    \148\001\139\001\157\001\150\001\159\001\161\001\157\001\166\001\
    \148\001\164\001\169\001\157\001\153\001\157\001\173\001\157\001\
    \171\001\164\001\159\001\179\001\164\001\162\001\180\001\180\001\
    \172\001\177\001\186\001\177\001\173\001\173\001\177\001\175\001\
    \195\001\181\001\187\001\185\001\185\001\204\001\189\001\193\001\
    \187\001\205\001\196\001\210\001\193\001\203\001\207\001\204\001\
    \209\001\199\001\219\001\203\001\218\001\055\002\056\002\217\001\
    \222\001\208\001\214\001\212\001\215\001\224\001\224\001\215\001\
    \222\001\231\001\221\001\232\001\219\001\071\002\072\002\222\001\
    \238\001\228\001\238\001\238\001\229\001\243\001\082\002\083\002\
    \247\001\248\001\254\001\242\001\238\001\242\001\002\002\242\001\
    \001\002\251\001\246\001\247\001\099\002\102\002\000\002\252\001\
    \007\002\003\002\000\002\012\002\008\002\016\002\111\002\112\002\
    \014\002\013\002\013\002\024\002\012\002\028\002\023\002\018\002\
    \020\002\020\002\031\002\019\002\039\002\027\002\027\002\044\002\
    \031\002\039\002\036\002\041\002\045\002\052\002\047\002\037\002\
    \148\002\149\002\040\002\053\002\042\002\051\002\060\002\053\002\
    \048\002\050\002\067\002\058\002\071\002\072\002\059\002\073\002\
    \075\002\065\002\167\002\169\002\074\002\059\002\069\002\089\002\
    \077\002\089\002\088\002\082\002\090\002\093\002\082\002\092\002\
    \091\002\099\002\097\002\088\002\196\002\101\002\099\002\097\002\
    \093\002\112\002\096\002\110\002\098\002\116\002\117\002\122\002\
    \110\002\122\002\121\002\124\002\108\002\108\002\126\002\126\002\
    \223\002\224\002\108\002\118\002\130\002\114\002\243\255\125\002\
    \123\002\135\002\127\002\123\002\148\002\148\002\148\002\149\002\
    \132\002\144\002\155\002\165\002\002\003\003\003\158\002\173\002\
    \171\002\160\002\176\002\158\002\171\002\162\002\163\002\178\002\
    \164\002\175\002\163\002\181\002\018\003\019\003\165\002\170\002\
    \173\002\191\002\190\002\176\002\188\002\183\002\185\002\182\002\
    \200\002\185\002\184\002\183\002\201\002\203\002\039\003\040\003\
    \198\002\194\002\214\002\196\002\208\002\203\002\205\002\219\002\
    \217\002\201\002\222\002\206\002\212\002\210\002\215\002\223\002\
    \228\002\227\002\213\002\225\002\220\002\222\002\236\002\220\002\
    \232\002\228\002\230\002\226\002\234\002\231\002\234\002\230\002\
    \243\002\247\002\234\002\249\002\236\002\237\002\245\002\002\003\
    \243\002\240\002\001\003\255\002\002\003\009\003\247\002\003\003\
    \254\002\000\003\252\002\244\255\001\003\015\003\010\003\008\003\
    \022\003\006\003\020\003\113\003\115\003\015\003\015\003\027\003\
    \011\003\021\003\029\003\016\003\016\003\036\003\024\003\036\003\
    \035\003\027\003\037\003\023\003\040\003\246\255\038\003\033\003\
    \036\003\036\003\037\003\051\003\050\003\055\003\043\003\039\003\
    \043\003\059\003\043\003\057\003\049\003\044\003\050\003\048\003\
    \062\003\055\003\049\003\061\003\068\003\161\003\162\003\057\003\
    \053\003\075\003\057\003\069\003\064\003\066\003\076\003\070\003\
    \075\003\066\003\083\003\064\003\086\003\070\003\070\003\082\003\
    \078\003\080\003\096\003\092\003\010\000\098\003\102\003\090\003\
    \086\003\088\003\086\003\092\003\102\003\092\003\108\003\109\003\
    \020\000\103\003\115\003\098\003\099\003\098\003\117\003\101\003\
    \101\003\105\003\119\003\120\003\113\003\122\003\106\003\124\003\
    \111\003\126\003\112\003\119\003\110\003\132\003\116\003\121\003\
    \123\003\130\003\126\003\121\003\139\003\123\003\136\003\131\003\
    \133\003\147\003\137\003\237\003\238\003\149\003\145\003\149\003\
    \142\003\151\003\153\003\140\003\150\003\146\003\155\003\143\003\
    \158\003\158\003\161\003\149\003\149\003\160\003\152\003\168\003\
    \163\003\165\003\160\003\156\003\013\004\247\255\014\004\165\003\
    \176\003\166\003\179\003\179\003\175\003\183\003\188\003\172\003\
    \188\003\190\003\012\000\180\003\181\003\181\003\197\003\186\003\
    \031\004\032\004\181\003\187\003\189\003\205\003\183\003\208\003\
    \209\003\196\003\213\003\211\003\208\003\203\003\216\003\232\003\
    \223\003\245\003\233\003\013\004\225\003\239\003\242\003\240\003\
    \234\003\232\003\246\003\239\003\239\003\236\003\240\003\010\004\
    \005\004\019\004\011\004\006\004\007\004\019\004\026\004\117\004\
    \119\004\018\004\019\004\039\004\021\004\033\004\028\004\030\004\
    \040\004\045\004\034\004\036\004\049\004\050\004\048\004\036\004\
    \052\004\044\004\035\004\057\004\041\004\046\004\048\004\055\004\
    \051\004\046\004\064\004\048\004\060\004\055\004\057\004\071\004\
    \061\004\161\004\162\004\073\004\069\004\073\004\066\004\075\004\
    \080\004\076\004\067\004\085\004\062\004\067\004\076\004\084\004\
    \177\004\178\004\077\004\093\004\091\004\086\004\102\004\084\004\
    \097\004\088\004\089\004\104\004\090\004\101\004\089\004\107\004\
    \202\004\203\004\093\004\098\004\101\004\119\004\118\004\102\004\
    \114\004\109\004\111\004\106\004\126\004\110\004\109\004\129\004\
    \117\004\131\004\163\004\115\004\125\004\122\004\132\004\127\004\
    \131\004\132\004\145\004\129\004\123\004\126\004\135\004\137\004\
    \144\004\140\004\135\004\152\004\136\004\156\004\140\004\152\004\
    \140\004\158\004\159\004\160\004\163\004\147\004\148\004\152\004\
    \166\004\148\004\155\004\175\004\170\004\165\004\168\004\164\004\
    \165\004\170\004\172\004\178\004\185\004\188\004\178\004\175\004\
    \189\004\182\004\174\004\190\004\178\004\193\004\180\004\190\004\
    \186\004\186\004\196\004\198\004\186\004\188\004\199\004\189\004\
    \201\004\196\004\198\004\194\004\190\004\213\004\197\004\213\004\
    \215\004\013\000\205\004\206\004\206\004\203\004\223\004\212\004\
    \221\004\207\004\223\004\211\004\211\004\216\004\231\004\221\004\
    \232\004\234\004\224\004\068\005\069\005\225\004\221\004\221\004\
    \241\004\229\004\241\004\240\004\231\004\244\004\236\004\240\004\
    \237\004\243\004\233\004\245\004\240\004\242\004\000\005\240\004\
    \254\004\093\005\094\005\249\004\251\004\247\004\247\004\011\005\
    \255\004\011\005\010\005\012\005\252\004\006\005\014\005\004\005\
    \020\005\014\005\006\005\023\005\023\000\014\005\026\005\009\005\
    \012\005\011\005\030\005\014\005\014\005\018\005\032\005\033\005\
    \026\005\035\005\023\005\022\005\034\005\029\005\031\005\044\005\
    \040\005\080\005\045\005\049\005\037\005\032\005\034\005\032\005\
    \038\005\048\005\038\005\054\005\059\005\057\005\041\005\062\005\
    \046\005\052\005\051\005\056\005\064\005\069\005\069\005\053\005\
    \065\005\061\005\063\005\077\005\059\005\071\005\066\005\069\005\
    \065\005\073\005\062\005\082\005\084\005\071\005\081\005\090\005\
    \090\005\081\005\074\005\090\005\077\005\081\005\093\005\096\005\
    \089\005\086\005\082\005\102\005\087\005\099\005\087\005\105\005\
    \106\005\111\005\099\005\111\005\111\005\114\005\098\005\098\005\
    \116\005\116\005\209\005\210\005\099\005\109\005\121\005\113\005\
    \108\005\120\005\110\005\106\005\120\005\124\005\131\005\126\005\
    \116\005\121\005\123\005\119\005\132\005\136\005\123\005\138\005\
    \125\005\129\005\134\005\146\005\142\005\137\005\140\005\135\005\
    \134\005\136\005\153\005\137\005\147\005\155\005\160\005\145\005\
    \142\005\159\005\156\005\146\005\147\005\167\005\156\005\165\005\
    \151\005\167\005\155\005\155\005\155\005\152\005\157\005\009\006\
    \010\006\178\005\169\005\173\005\176\005\183\005\165\005\177\005\
    \172\005\174\005\188\005\170\005\182\005\177\005\179\005\185\005\
    \191\005\176\005\188\005\184\005\196\005\194\005\184\005\185\005\
    \201\005\203\005\190\005\190\005\192\005\198\005\201\005\196\005\
    \192\005\212\005\214\005\201\005\211\005\220\005\220\005\211\005\
    \204\005\220\005\207\005\223\005\207\005\217\005\225\005\062\006\
    \242\255\064\006\221\005\236\005\223\005\238\005\239\005\237\005\
    \243\005\235\005\231\005\250\005\000\006\255\005\012\006\010\006\
    \031\006\040\006\241\005\015\006\043\006\048\006\251\255\049\000\
    \152\006\083\000\001\000\146\006\157\006\085\000\252\255\054\000\
    \159\006\254\255\010\006\247\005\074\006\248\005\010\006\066\006\
    \016\006\147\006\148\006\008\006\036\006\049\006\074\006\068\006\
    \065\006\076\006\063\006\084\006\076\006\077\006\076\006\082\006\
    \080\006\090\006\097\006\092\006\082\006\190\006\191\006\087\006\
    \099\006\086\006\095\006\104\006\096\006\091\006\093\006\250\255\
    \093\006\113\006\111\006\116\006\106\006\209\006\210\006\109\006\
    \113\006\111\006\124\006\120\006\134\006\124\006\126\006\124\006\
    \124\006\133\006\128\006\237\006\238\006\137\006\139\006\139\006\
    \138\006\144\006\153\006\135\006\141\006\157\006\145\006\150\006\
    \146\006\151\006\149\006\159\006\150\006\166\006\160\006\153\006\
    \171\006\164\006\173\006\175\006\171\006\180\006\162\006\178\006\
    \183\006\171\006\167\006\171\006\187\006\171\006\186\006\178\006\
    \173\006\193\006\178\006\175\006\192\006\192\006\184\006\188\006\
    \197\006\188\006\184\006\184\006\187\006\185\006\204\006\188\006\
    \195\006\194\006\194\006\213\006\199\006\203\006\198\006\215\006\
    \208\006\206\006\223\006\227\006\210\006\220\006\224\006\221\006\
    \226\006\216\006\236\006\220\006\235\006\072\007\073\007\234\006\
    \239\006\225\006\231\006\229\006\232\006\241\006\241\006\232\006\
    \239\006\248\006\238\006\249\006\236\006\088\007\089\007\239\006\
    \255\006\245\006\255\006\255\006\246\006\004\007\099\007\100\007\
    \008\007\009\007\015\007\003\007\255\006\003\007\019\007\003\007\
    \018\007\012\007\007\007\008\007\116\007\119\007\017\007\013\007\
    \024\007\020\007\017\007\029\007\025\007\033\007\128\007\129\007\
    \031\007\030\007\030\007\041\007\029\007\045\007\040\007\035\007\
    \037\007\037\007\048\007\036\007\056\007\044\007\044\007\061\007\
    \048\007\056\007\053\007\058\007\062\007\069\007\064\007\054\007\
    \165\007\166\007\057\007\070\007\059\007\068\007\077\007\070\007\
    \065\007\067\007\068\007\083\007\069\007\247\255\087\007\078\007\
    \091\007\092\007\079\007\093\007\096\007\087\007\191\007\192\007\
    \093\007\089\007\089\007\109\007\097\007\109\007\108\007\102\007\
    \110\007\113\007\104\007\109\007\113\007\120\007\116\007\106\007\
    \216\007\121\007\119\007\117\007\113\007\133\007\118\007\130\007\
    \118\007\136\007\137\007\142\007\130\007\142\007\141\007\144\007\
    \128\007\128\007\146\007\146\007\240\007\241\007\130\007\140\007\
    \152\007\134\007\246\255\145\007\140\007\152\007\142\007\141\007\
    \170\007\162\007\169\007\170\007\148\007\156\007\175\007\183\007\
    \022\008\023\008\178\007\193\007\191\007\180\007\196\007\178\007\
    \191\007\182\007\183\007\198\007\184\007\195\007\183\007\201\007\
    \038\008\039\008\185\007\190\007\193\007\211\007\210\007\196\007\
    \208\007\203\007\205\007\202\007\220\007\205\007\204\007\203\007\
    \221\007\223\007\059\008\063\008\218\007\214\007\234\007\216\007\
    \228\007\223\007\225\007\239\007\237\007\221\007\242\007\226\007\
    \232\007\230\007\235\007\243\007\248\007\247\007\232\007\244\007\
    \239\007\242\007\000\008\240\007\252\007\248\007\250\007\246\007\
    \254\007\251\007\254\007\250\007\007\008\011\008\254\007\013\008\
    \000\008\001\008\009\008\022\008\007\008\004\008\021\008\019\008\
    \022\008\029\008\011\008\023\008\018\008\020\008\016\008\245\255\
    \035\008\023\008\064\008\038\008\023\008\024\008\026\008\036\008\
    \033\008\039\008\034\008\038\008\039\008\051\008\036\008\030\008\
    \033\008\042\008\044\008\051\008\047\008\042\008\059\008\043\008\
    \062\008\046\008\058\008\046\008\065\008\066\008\067\008\070\008\
    \054\008\054\008\058\008\072\008\055\008\060\008\078\008\074\008\
    \069\008\071\008\071\008\073\008\079\008\086\008\089\008\079\008\
    \076\008\090\008\083\008\075\008\091\008\079\008\083\008\097\008\
    \092\008\090\008\103\008\087\008\101\008\194\008\195\008\096\008\
    \096\008\108\008\092\008\102\008\110\008\097\008\097\008\117\008\
    \105\008\117\008\116\008\108\008\118\008\104\008\121\008\248\255\
    \119\008\114\008\118\008\117\008\118\008\132\008\131\008\136\008\
    \124\008\120\008\124\008\140\008\124\008\138\008\130\008\125\008\
    \131\008\129\008\143\008\136\008\130\008\142\008\149\008\241\008\
    \242\008\138\008\136\008\156\008\138\008\150\008\145\008\147\008\
    \157\008\151\008\156\008\147\008\164\008\145\008\167\008\151\008\
    \151\008\164\008\159\008\165\008\177\008\175\008\215\008\180\008\
    \184\008\173\008\168\008\170\008\168\008\174\008\184\008\174\008\
    \190\008\191\008\226\008\186\008\198\008\181\008\183\008\182\008\
    \201\008\185\008\185\008\189\008\203\008\204\008\197\008\206\008\
    \190\008\208\008\194\008\191\008\196\008\048\009\049\009\217\008\
    \208\008\212\008\215\008\222\008\205\008\217\008\212\008\214\008\
    \249\255\228\008\210\008\222\008\217\008\219\008\230\008\217\008\
    \227\008\223\008\231\008\221\008\238\008\239\008\239\008\079\009\
    \080\009\231\008\243\008\233\008\247\008\247\008\243\008\251\008\
    \241\008\241\008\252\008\242\008\254\008\249\008\251\008\246\008\
    \242\008\009\009\249\008\009\009\013\009\038\009\004\009\005\009\
    \005\009\021\009\010\009\111\009\135\009\004\009\010\009\012\009\
    \029\009\003\009\030\009\034\009\021\009\028\009\034\009\033\009\
    \034\009\039\009\055\009\036\009\057\009\059\009\064\009\044\009\
    \058\009\067\009\064\009\058\009\057\009\071\009\064\009\065\009\
    \061\009\065\009\082\009\066\009\080\009\072\009\067\009\068\009\
    \081\009\088\009\181\009\200\009\077\009\073\009\094\009\076\009\
    \088\009\083\009\085\009\109\009\123\009\114\009\115\009\128\009\
    \129\009\127\009\112\009\128\009\121\009\113\009\135\009\119\009\
    \124\009\126\009\133\009\129\009\124\009\142\009\126\009\138\009\
    \133\009\135\009\149\009\139\009\239\009\240\009\151\009\147\009\
    \151\009\144\009\153\009\158\009\154\009\145\009\163\009\140\009\
    \145\009\154\009\162\009\255\009\000\010\155\009\171\009\169\009\
    \163\009\179\009\161\009\175\009\166\009\167\009\182\009\168\009\
    \179\009\167\009\185\009\024\010\025\010\171\009\176\009\179\009\
    \197\009\196\009\180\009\192\009\187\009\189\009\184\009\204\009\
    \188\009\187\009\207\009\195\009\209\009\237\009\195\009\205\009\
    \204\009\210\009\205\009\209\009\210\009\222\009\206\009\201\009\
    \204\009\213\009\215\009\222\009\218\009\213\009\230\009\214\009\
    \234\009\218\009\232\009\221\009\239\009\240\009\241\009\244\009\
    \228\009\228\009\232\009\247\009\229\009\236\009\254\009\249\009\
    \244\009\246\009\242\009\243\009\248\009\250\009\000\010\007\010\
    \010\010\000\010\253\009\011\010\004\010\252\009\012\010\000\010\
    \015\010\002\010\012\010\008\010\008\010\018\010\020\010\008\010\
    \010\010\021\010\011\010\023\010\018\010\020\010\016\010\012\010\
    \035\010\019\010\035\010\037\010\062\010\028\010\029\010\029\010\
    \026\010\046\010\035\010\044\010\030\010\046\010\034\010\034\010\
    \039\010\054\010\044\010\055\010\057\010\047\010\147\010\148\010\
    \048\010\044\010\044\010\064\010\052\010\064\010\063\010\054\010\
    \067\010\059\010\063\010\060\010\066\010\056\010\068\010\063\010\
    \065\010\079\010\063\010\077\010\172\010\173\010\072\010\074\010\
    \070\010\070\010\090\010\078\010\090\010\089\010\091\010\075\010\
    \085\010\093\010\083\010\099\010\093\010\085\010\102\010\134\010\
    \094\010\106\010\091\010\092\010\091\010\110\010\094\010\094\010\
    \098\010\112\010\113\010\106\010\115\010\103\010\102\010\114\010\
    \109\010\112\010\124\010\120\010\160\010\125\010\129\010\117\010\
    \112\010\114\010\112\010\118\010\128\010\118\010\134\010\139\010\
    \137\010\121\010\142\010\126\010\132\010\131\010\136\010\144\010\
    \150\010\149\010\133\010\146\010\141\010\143\010\157\010\139\010\
    \151\010\146\010\149\010\145\010\153\010\142\010\162\010\164\010\
    \151\010\161\010\170\010\170\010\161\010\154\010\170\010\157\010\
    \161\010\173\010\176\010\169\010\166\010\162\010\182\010\167\010\
    \179\010\167\010\185\010\186\010\191\010\179\010\191\010\191\010\
    \194\010\178\010\178\010\196\010\196\010\033\011\034\011\179\010\
    \189\010\201\010\193\010\188\010\200\010\190\010\186\010\200\010\
    \204\010\211\010\206\010\196\010\201\010\203\010\199\010\212\010\
    \216\010\203\010\218\010\205\010\209\010\214\010\226\010\222\010\
    \217\010\220\010\215\010\214\010\216\010\233\010\217\010\227\010\
    \235\010\240\010\225\010\222\010\239\010\236\010\226\010\227\010\
    \247\010\236\010\245\010\231\010\247\010\235\010\235\010\246\010\
    \252\010\237\010\249\010\244\010\000\011\252\010\242\010\243\010\
    \003\011\005\011\249\010\246\010\248\010\000\011\003\011\254\010\
    \004\011\251\010\017\011\001\011\007\011\009\011\016\011\012\011\
    \007\011\025\011\009\011\021\011\016\011\018\011\032\011\022\011\
    \122\011\123\011\034\011\030\011\034\011\027\011\036\011\018\011\
    \038\011\040\011\027\011\037\011\046\011\046\011\037\011\030\011\
    \046\011\033\011\049\011\033\011\043\011\051\011\148\011\249\255\
    \250\255\251\255\093\011\114\011\153\011\155\011\160\011\162\011\
    \164\011\169\011\173\011\203\011\182\011\188\011\167\011\185\011\
    \174\011\255\255\212\011\201\011\219\011\249\255\208\011\150\011\
    \076\011\110\011\255\255\228\011\210\011\221\011\226\011\253\255\
    \169\011\252\255\171\011\214\011\250\255\251\255\238\011\009\012\
    \233\255\234\255\202\012\093\011\094\011\189\011\168\011\211\011\
    \214\011\213\011\216\011\157\011\174\011\161\011\006\012\030\013\
    \255\255\239\255\252\255\058\012\000\012\254\255\251\011\253\255\
    \252\011\251\255\249\255\124\012\235\011\226\011\126\012\131\012\
    \133\012\144\012\229\011\248\255\244\255\242\255\240\255\009\012\
    \237\255\245\011\252\255\241\011\254\255\255\255\253\255\248\011\
    \252\255\245\011\254\255\255\255\253\255\153\012\250\255\247\011\
    \252\255\254\011\000\012\253\255\254\255\251\255\159\012\249\255\
    \250\255\251\011\251\011\232\011\233\011\030\012\236\011\241\011\
    \168\012\255\255\057\012\170\012\253\255\252\255\251\255\234\011\
    \253\255\201\011\255\255\236\011\254\255\255\255\145\013\248\255\
    \177\012\064\012\142\013\209\013\018\014\143\011\171\014\251\255\
    \097\012\103\012\254\255\255\255\253\255\252\255\229\014\083\012\
    \055\015\102\015\166\015\218\011\032\013\230\015\012\013\140\013\
    \093\012\021\016\068\016\132\016\221\011\251\014\196\016\176\014\
    \213\013\123\017\251\255\201\012\183\012\126\012\128\017\133\017\
    \254\255\184\012\252\255\200\012\251\255\252\255\110\012\110\012\
    \255\255\061\013\254\255\145\012\202\014\135\017\241\255\242\255\
    \149\012\166\012\244\255\126\012\125\012\131\012\077\012\129\012\
    \140\012\255\255\107\015\254\255\179\012\133\015\194\012\251\255\
    \182\012\249\255\247\255\245\255\203\012\252\255\163\012\254\255\
    \138\015\255\255\196\012\140\015\253\255\141\017\249\255\215\012\
    \145\015\178\012\254\255\150\015\255\255\217\012\165\015\253\255\
    \252\255\220\012\169\017\126\017\254\255\228\012\204\017\241\255\
    \242\255\243\255\199\012\182\017\012\018\077\018\046\012\230\018\
    \137\019\190\012\089\019\252\255\142\012\175\017\229\012\151\012\
    \148\013\202\017\236\012\184\017\254\255\251\012\186\017\253\255\
    \252\012\244\255\251\255\170\012\254\012\112\018\201\012\193\017\
    \250\255\030\013\000\013\195\017\248\255\195\019\215\012\021\020\
    \068\020\132\020\148\012\029\021\192\021\099\022\006\023\008\019\
    \024\021\007\015\123\017\049\013\170\021\077\022\001\023\161\012\
    \169\023\075\013\147\023\233\023\041\024\193\012\024\019\105\024\
    \202\018\190\019\218\019\169\024\188\019\205\019\019\020\233\024\
    \200\019\007\020\115\018\253\255\254\255\255\255\195\014\253\255\
    \090\013\255\255\196\014\253\255\122\013\255\255\169\013\254\255\
    \134\013\255\255\131\017\076\013\106\013\144\017\157\017\132\013\
    \139\013\177\013\177\017\238\013\211\017\014\014\171\014\254\255\
    \255\255\240\013\016\014\017\014\097\022\253\255\191\013\255\255\
    \068\014\232\017\069\014\121\018\243\013\252\014\254\255\071\014\
    \160\025\254\255\136\017\146\017\099\022\006\023\161\025\255\255\
    \151\017\119\018\166\025\164\019\098\022\255\255\017\022\168\025\
    \005\023\192\023\254\255\170\025\172\025\174\025\161\014\254\255\
    \040\014\055\014\191\017\104\022\159\017\230\017\167\025\254\255\
    \238\014\145\025\030\026\094\026\029\014\012\027\231\017\163\025\
    \070\027\234\027\141\018\212\027\042\028\106\028\031\014\148\022\
    \170\028\194\021\211\023\159\018\217\028\008\029\072\029\033\014\
    \246\029\227\018\224\029\054\030\118\030\046\014\209\025\182\030\
    \173\023\237\023\247\025\246\030\031\025\094\025\010\026\054\031\
    \110\026\235\025\137\017\255\255\034\018\108\018";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \000\000\019\000\019\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\017\000\255\255\255\255\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\014\000\255\255\013\000\013\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\016\000\255\255\255\255\
    \255\255\015\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\011\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\010\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\009\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\011\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\011\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \011\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\011\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\011\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\011\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\255\255\013\000\
    \000\000\013\000\255\255\255\255\255\255\255\255\255\255\002\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \008\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\008\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\010\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\010\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\008\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\006\000\001\000\002\000\006\000\006\000\006\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\003\000\255\255\255\255\006\000\006\000\
    \006\000\006\000\255\255\006\000\001\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000\255\255\
    \255\255\255\255\022\000\022\000\020\000\022\000\014\000\012\000\
    \010\000\022\000\022\000\022\000\005\000\000\000\022\000\017\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\009\000\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\019\000\
    \255\255\255\255\255\255\003\000\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\255\255\255\255\255\255\005\000\
    \255\255\005\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\006\000\006\000\006\000\006\000\006\000\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\006\000\255\255\
    \006\000\007\000\007\000\007\000\007\000\007\000\005\000\255\255\
    \007\000\007\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\002\000\005\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\004\000\004\000\
    \255\255\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \014\000\012\000\255\255\009\000\007\000\005\000\003\000\014\000\
    \014\000\255\255\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\003\000\255\255\
    \003\000\255\255\255\255\001\000\255\255\005\000\255\255\005\000\
    \004\000\006\000\255\255\006\000\255\255\255\255\001\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\014\000\014\000\014\000\014\000\014\000\010\000\
    \010\000\014\000\014\000\255\255\014\000\000\000\014\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\004\000\255\255\006\000\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\010\000\010\000\009\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\001\000\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\001\000\000\000\255\255\255\255\255\255\002\000\255\255\
    \002\000\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \001\000\001\000\001\000\001\000\001\000\001\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\001\000\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\068\000\068\000\068\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\086\000\086\000\086\000\
    \255\255\086\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\069\002\000\000\069\002\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\136\004\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\111\006\
    \111\006\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\175\008\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\192\008\255\255\255\255\255\255\192\008\
    \192\008\000\000\255\255\255\255\197\008\000\000\212\008\255\255\
    \255\255\255\255\000\000\197\008\255\255\255\255\255\255\000\000\
    \255\255\000\000\255\255\212\008\000\000\000\000\212\008\219\008\
    \000\000\000\000\000\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\233\008\
    \000\000\000\000\000\000\255\255\255\255\000\000\255\255\000\000\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\255\255\
    \000\000\002\009\000\000\255\255\000\000\000\000\000\000\008\009\
    \000\000\255\255\000\000\000\000\000\000\014\009\000\000\255\255\
    \000\000\255\255\255\255\000\000\000\000\000\000\023\009\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\000\000\000\000\000\000\040\009\
    \000\000\255\255\000\000\044\009\000\000\000\000\047\009\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\084\009\000\000\084\009\084\009\089\009\084\009\084\009\
    \000\000\089\009\000\000\092\009\000\000\000\000\255\255\255\255\
    \000\000\255\255\000\000\255\255\255\255\102\009\000\000\000\000\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\000\000\125\009\000\000\255\255\000\000\
    \255\255\000\000\255\255\255\255\000\000\134\009\000\000\255\255\
    \255\255\255\255\000\000\255\255\000\000\255\255\255\255\000\000\
    \000\000\255\255\255\255\148\009\000\000\255\255\151\009\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\000\000\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\227\009\000\000\000\000\000\000\231\009\000\000\
    \255\255\000\000\235\009\000\000\255\255\000\000\239\009\000\000\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\009\000\000\
    \000\000\255\255\255\255\255\255\005\010\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \018\010\000\000\018\010\018\010\018\010\018\010\024\010\000\000\
    \024\010\024\010\027\010\027\010\027\010\000\000\027\010\027\010\
    \027\010\027\010\000\000\027\010\027\010\027\010\039\010\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\047\010\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\092\010\000\000\093\010\093\010";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\026\000\024\000\024\000\024\000\025\000\048\000\024\000\
    \056\000\055\000\067\000\104\000\055\000\069\000\070\000\069\000\
    \070\000\076\000\085\000\085\000\088\000\111\000\088\000\088\000\
    \026\000\096\000\103\000\104\000\115\000\122\000\105\000\056\000\
    \023\000\067\000\050\000\050\000\069\000\070\000\069\000\070\000\
    \076\000\085\000\085\000\088\000\091\000\088\000\088\000\103\000\
    \096\000\103\000\110\000\115\000\087\000\087\000\089\000\087\000\
    \090\000\006\000\005\000\022\000\014\000\017\000\010\000\020\000\
    \016\000\019\000\121\000\118\002\021\000\003\000\008\000\015\000\
    \009\000\007\000\018\000\012\000\013\000\011\000\004\000\129\002\
    \203\002\146\003\205\003\167\004\117\002\160\004\166\004\164\004\
    \168\004\123\004\107\004\171\002\108\004\104\004\149\002\085\004\
    \107\002\204\001\104\002\123\004\108\002\204\001\148\002\131\004\
    \120\004\084\004\144\002\045\002\172\002\103\004\103\002\075\002\
    \106\004\170\002\147\002\105\002\106\002\030\002\076\002\000\002\
    \044\002\243\000\001\002\164\001\131\004\120\004\144\002\255\001\
    \077\002\002\002\190\000\031\002\166\001\165\001\104\004\074\002\
    \150\001\021\002\113\000\149\001\192\000\126\000\049\000\024\000\
    \125\000\191\000\048\000\107\004\124\000\108\004\103\004\085\004\
    \127\000\171\002\149\002\128\000\150\001\092\000\045\002\149\001\
    \243\000\084\004\148\002\030\002\107\002\049\000\104\002\071\000\
    \108\002\106\004\172\002\044\002\075\002\047\000\147\002\170\002\
    \021\002\031\002\103\002\076\002\060\000\061\000\000\002\105\002\
    \106\002\001\002\062\000\063\000\064\000\077\002\255\001\065\000\
    \002\002\066\000\072\000\073\000\074\002\074\000\030\000\029\000\
    \046\000\038\000\041\000\034\000\044\000\040\000\043\000\075\000\
    \078\000\045\000\027\000\032\000\039\000\033\000\031\000\042\000\
    \036\000\037\000\035\000\028\000\164\001\190\000\049\000\024\000\
    \053\000\079\000\048\000\059\000\053\000\166\001\165\001\192\000\
    \126\000\059\000\080\000\125\000\191\000\081\000\082\000\124\000\
    \076\000\083\000\084\000\127\000\093\000\049\000\128\000\053\000\
    \002\000\094\000\059\000\053\000\095\000\047\000\088\000\106\000\
    \059\000\054\000\099\000\100\000\255\255\255\255\255\255\076\000\
    \101\000\096\000\255\255\255\255\107\000\255\255\108\000\114\000\
    \117\000\118\000\119\000\178\000\169\000\088\000\030\000\029\000\
    \046\000\038\000\041\000\034\000\044\000\040\000\043\000\087\000\
    \096\000\045\000\027\000\032\000\039\000\033\000\031\000\042\000\
    \036\000\037\000\035\000\028\000\101\000\108\000\110\000\111\000\
    \115\000\184\000\112\000\173\000\119\000\185\000\153\000\147\000\
    \141\000\052\000\121\000\122\000\051\000\052\000\123\000\132\000\
    \051\000\174\000\131\000\101\000\108\000\110\000\135\000\115\000\
    \136\000\161\000\058\000\119\000\057\000\102\000\109\000\137\000\
    \058\000\121\000\057\000\160\000\138\000\120\000\139\000\077\000\
    \140\000\140\000\142\000\143\000\140\000\144\000\129\000\145\000\
    \146\000\148\000\149\000\150\000\151\000\152\000\134\000\130\000\
    \098\000\125\000\154\000\133\000\155\000\156\000\157\000\140\000\
    \158\000\159\000\125\000\097\000\167\000\162\000\163\000\164\000\
    \165\000\166\000\125\000\168\000\125\000\170\000\171\000\172\000\
    \125\000\177\000\175\000\176\000\125\000\179\000\180\000\181\000\
    \182\000\183\000\125\000\187\000\186\000\125\000\188\000\189\000\
    \125\000\147\001\146\001\193\000\176\000\194\000\195\000\197\000\
    \125\000\125\000\116\000\198\000\198\000\099\001\090\001\196\000\
    \051\001\050\001\045\001\243\000\239\000\234\000\227\000\209\000\
    \210\000\212\000\226\000\191\000\211\000\213\000\214\000\215\000\
    \216\000\217\000\198\000\198\000\217\000\219\000\220\000\221\000\
    \222\000\223\000\224\000\225\000\176\000\176\000\228\000\176\000\
    \229\000\230\000\231\000\232\000\191\000\235\000\232\000\236\000\
    \217\000\237\000\238\000\217\000\176\000\240\000\241\000\242\000\
    \176\000\244\000\245\000\246\000\246\000\035\001\202\000\201\000\
    \205\000\031\001\232\000\204\000\208\000\232\000\027\001\207\000\
    \018\001\011\001\203\000\002\001\254\000\206\000\199\000\200\000\
    \035\000\255\000\246\000\246\000\218\000\000\001\001\001\176\000\
    \003\001\008\001\005\001\006\001\004\001\255\255\007\001\176\000\
    \009\001\010\001\176\000\012\001\013\001\014\001\015\001\016\001\
    \017\001\176\000\019\001\020\001\021\001\022\001\023\001\024\001\
    \025\001\233\000\026\001\176\000\028\001\029\001\247\000\030\001\
    \253\000\176\000\032\001\033\001\251\000\034\001\176\000\036\001\
    \037\001\038\001\249\000\039\001\040\001\252\000\041\001\250\000\
    \248\000\042\001\043\001\044\001\176\000\046\001\048\001\176\000\
    \049\001\047\001\176\000\174\000\052\001\053\001\085\001\058\001\
    \055\001\056\001\057\001\176\000\059\001\060\001\054\001\061\001\
    \062\001\062\001\082\001\075\001\073\001\067\001\068\001\069\001\
    \070\001\071\001\072\001\176\000\074\001\176\000\076\001\077\001\
    \078\001\078\001\080\001\081\001\176\000\083\001\084\001\062\001\
    \062\001\077\001\086\001\087\001\087\001\089\001\176\000\091\001\
    \092\001\093\001\094\001\095\001\096\001\176\000\097\001\078\001\
    \078\001\098\001\176\000\100\001\101\001\130\001\104\001\101\001\
    \105\001\106\001\087\001\087\001\107\001\108\001\109\001\110\001\
    \111\001\111\001\129\001\122\001\115\001\116\001\117\001\118\001\
    \119\001\120\001\121\001\101\001\064\001\066\001\101\001\176\000\
    \123\001\124\001\063\001\065\001\125\001\126\001\127\001\111\001\
    \111\001\128\001\176\000\176\000\131\001\132\001\133\001\134\001\
    \135\001\136\001\139\001\079\001\140\001\137\001\137\001\141\001\
    \142\001\143\001\088\001\144\001\145\001\176\000\176\000\148\001\
    \146\000\163\001\151\001\152\001\153\001\154\001\158\001\102\001\
    \155\001\112\001\155\001\114\001\137\001\137\001\103\001\209\000\
    \157\001\159\001\160\001\161\001\162\001\125\000\146\000\199\001\
    \174\001\176\001\167\001\173\001\169\001\113\001\170\001\155\001\
    \168\001\155\001\171\001\211\000\172\001\217\000\197\001\189\001\
    \175\001\182\001\177\001\178\001\179\001\183\001\138\001\180\001\
    \181\001\125\000\187\001\184\001\185\001\186\001\125\000\188\001\
    \125\000\190\001\191\001\192\001\217\000\195\001\196\001\125\000\
    \193\001\193\001\198\001\200\001\202\000\201\000\205\000\201\001\
    \202\001\204\000\156\001\203\001\125\000\207\000\209\001\206\001\
    \203\000\239\001\211\001\206\000\199\000\200\000\035\000\193\001\
    \193\001\207\001\210\001\222\001\208\001\235\000\146\000\205\001\
    \240\001\221\001\212\001\213\001\213\001\215\001\216\001\217\001\
    \218\001\219\001\220\001\125\000\224\001\223\001\125\000\225\001\
    \226\001\227\001\228\001\229\001\229\001\231\001\232\001\233\001\
    \234\001\235\001\213\001\213\001\236\001\237\001\238\001\125\000\
    \243\001\241\001\125\000\242\001\125\000\244\001\245\001\246\001\
    \247\001\247\001\229\001\229\001\194\001\249\001\250\001\251\001\
    \252\001\253\001\254\001\198\001\020\002\125\000\014\002\004\002\
    \008\002\005\002\006\002\007\002\125\000\009\002\010\002\247\001\
    \247\001\011\002\012\002\013\002\125\000\015\002\240\000\214\001\
    \016\002\017\002\003\002\018\002\019\002\125\000\125\000\022\002\
    \230\001\023\002\024\002\025\002\026\002\027\002\028\002\125\000\
    \125\000\035\002\029\002\032\002\033\002\034\002\176\000\047\001\
    \036\002\037\002\038\002\039\002\040\002\041\002\042\002\043\002\
    \073\002\052\001\047\002\065\002\048\002\046\002\248\001\049\002\
    \050\002\051\002\052\002\054\001\052\002\059\002\055\002\056\002\
    \057\002\058\002\125\000\060\002\061\002\062\002\063\002\064\002\
    \125\000\066\002\068\002\070\002\069\002\068\001\067\002\071\002\
    \072\002\052\002\125\000\052\002\125\000\102\002\090\002\086\002\
    \078\002\079\002\080\002\081\002\082\002\083\002\084\002\085\002\
    \176\000\087\002\088\002\089\002\146\000\091\002\092\002\093\002\
    \096\002\097\002\094\002\094\002\098\002\099\002\100\002\101\002\
    \125\000\125\000\142\002\139\002\133\002\128\002\127\002\109\002\
    \114\002\110\002\111\002\112\002\053\002\113\002\125\000\064\001\
    \054\002\094\002\094\002\115\002\116\002\063\001\065\001\122\002\
    \119\002\120\002\100\001\121\002\125\000\123\002\124\002\125\002\
    \126\002\125\000\069\002\130\002\131\002\132\002\125\000\134\002\
    \135\002\136\002\137\002\138\002\125\000\140\002\141\002\125\000\
    \143\002\043\002\145\002\146\002\100\001\156\002\151\002\150\002\
    \146\000\152\002\153\002\154\002\155\002\125\000\157\002\158\002\
    \095\002\159\002\160\002\161\002\162\002\163\002\164\002\164\002\
    \166\002\167\002\168\002\169\002\125\000\125\000\077\004\173\002\
    \076\004\206\002\175\002\198\002\188\002\174\002\179\002\186\002\
    \181\002\182\002\177\002\178\002\183\002\164\002\164\002\176\002\
    \180\002\184\002\185\002\146\000\187\002\146\000\190\002\190\002\
    \195\002\193\002\194\002\189\002\196\002\197\002\189\002\199\002\
    \200\002\201\002\202\002\204\002\205\002\125\000\207\002\208\002\
    \209\002\209\002\074\004\070\004\066\004\190\002\190\002\058\004\
    \049\004\059\004\005\004\002\004\014\004\173\003\201\003\157\003\
    \200\003\126\003\158\003\189\002\189\002\013\004\015\004\209\002\
    \209\002\165\002\172\003\237\003\032\003\060\004\238\003\202\003\
    \203\003\072\003\127\003\236\003\043\002\239\003\002\003\125\003\
    \071\003\000\003\244\002\236\002\073\003\232\002\011\003\074\003\
    \001\003\233\002\234\002\235\002\043\002\237\002\010\003\238\002\
    \239\002\213\002\217\002\223\002\220\002\218\002\225\002\221\002\
    \214\002\215\002\009\003\240\002\226\002\212\002\229\002\211\002\
    \222\002\210\002\224\002\219\002\227\002\231\002\216\002\241\002\
    \242\002\243\002\043\002\245\002\246\002\247\002\248\002\230\002\
    \248\002\250\002\255\255\191\002\255\255\192\002\228\002\251\002\
    \252\002\253\002\254\002\255\002\043\002\043\002\006\003\003\003\
    \005\003\004\003\043\002\007\003\043\002\248\002\043\002\248\002\
    \008\003\043\002\018\003\013\003\012\003\043\002\014\003\015\003\
    \016\003\017\003\043\002\019\003\020\003\021\003\022\003\023\003\
    \024\003\025\003\026\003\026\003\028\003\029\003\030\003\031\003\
    \043\002\067\003\050\003\034\003\039\003\038\003\037\003\043\002\
    \049\003\040\003\041\003\041\003\043\003\035\003\044\003\045\003\
    \068\003\026\003\026\003\033\003\036\003\249\002\046\003\047\003\
    \048\003\043\002\052\003\051\003\043\002\053\003\054\003\055\003\
    \056\003\041\003\041\003\057\003\057\003\059\003\060\003\061\003\
    \062\003\063\003\064\003\065\003\066\003\043\002\043\002\069\003\
    \070\003\043\002\119\003\113\003\111\003\107\003\120\003\099\003\
    \091\003\078\003\057\003\057\003\077\003\085\003\080\003\081\003\
    \082\003\098\003\083\003\084\003\043\002\027\003\086\003\087\003\
    \088\003\089\003\090\003\043\002\092\003\093\003\042\003\094\003\
    \095\003\096\003\097\003\043\002\105\003\100\003\101\003\102\003\
    \075\003\103\003\104\003\043\002\106\003\255\255\255\255\043\002\
    \058\003\076\003\108\003\109\003\110\003\079\003\043\002\112\003\
    \043\002\114\003\115\003\116\003\117\003\118\003\043\002\122\003\
    \121\003\043\002\123\003\124\003\043\002\043\002\150\003\128\003\
    \149\003\141\003\131\003\132\003\139\003\129\003\130\003\134\003\
    \135\003\136\003\137\003\138\003\043\002\133\003\140\003\043\002\
    \142\003\143\003\144\003\145\003\147\003\148\003\043\002\043\002\
    \151\003\152\003\153\003\154\003\155\003\156\003\043\002\160\003\
    \159\003\043\002\161\003\162\003\163\003\164\003\164\003\166\003\
    \167\003\168\003\169\003\170\003\171\003\043\002\199\003\174\003\
    \182\003\176\003\175\003\177\003\178\003\179\003\180\003\181\003\
    \043\002\183\003\184\003\185\003\164\003\164\003\186\003\186\003\
    \195\003\189\003\190\003\191\003\192\003\193\003\194\003\043\002\
    \196\003\197\003\198\003\043\002\043\002\218\003\215\003\219\003\
    \209\003\204\003\206\003\207\003\208\003\186\003\186\003\043\002\
    \210\003\211\003\212\003\213\003\214\003\043\002\216\003\217\003\
    \043\002\223\003\220\003\221\003\222\003\043\002\165\003\224\003\
    \225\003\227\003\231\003\228\003\229\003\230\003\043\002\232\003\
    \233\003\234\003\235\003\043\002\001\004\043\002\251\003\241\003\
    \245\003\242\003\226\003\243\003\244\003\043\002\246\003\188\003\
    \247\003\248\003\249\003\187\003\250\003\043\002\252\003\253\003\
    \254\003\255\003\240\003\000\004\043\002\043\002\003\004\004\004\
    \043\002\006\004\007\004\008\004\009\004\011\004\043\002\012\004\
    \043\002\044\004\039\004\016\004\018\004\031\004\024\004\019\004\
    \010\004\020\004\025\004\021\004\022\004\023\004\043\002\029\004\
    \026\004\027\004\028\004\017\004\043\002\030\004\043\002\032\004\
    \033\004\034\004\035\004\035\004\037\004\038\004\043\002\040\004\
    \041\004\042\004\043\004\043\002\045\004\046\004\047\004\048\004\
    \043\002\050\004\051\004\052\004\053\004\054\004\055\004\056\004\
    \043\002\035\004\035\004\057\004\043\002\064\004\043\002\061\004\
    \062\004\063\004\043\002\065\004\043\002\067\004\068\004\069\004\
    \043\002\071\004\072\004\073\004\043\002\075\004\043\002\125\000\
    \078\004\079\004\080\004\081\004\082\004\083\004\125\000\098\004\
    \086\004\087\004\088\004\088\004\090\004\091\004\092\004\093\004\
    \094\004\095\004\096\004\097\004\189\002\099\004\100\004\101\004\
    \102\004\198\001\105\004\069\002\125\000\117\004\036\004\114\004\
    \125\000\088\004\088\004\110\004\111\004\112\004\113\004\069\002\
    \116\004\109\004\125\000\115\004\069\002\118\004\119\004\125\000\
    \121\004\122\004\125\000\124\004\125\004\126\004\127\004\129\004\
    \125\000\130\004\125\000\132\004\133\004\134\004\125\000\137\004\
    \160\004\137\004\128\004\161\004\170\008\162\008\159\008\138\008\
    \187\006\123\008\120\008\124\008\207\006\130\005\192\005\137\008\
    \134\005\089\004\186\006\149\006\207\004\146\006\137\004\150\006\
    \137\004\117\006\119\008\136\008\087\006\208\006\159\004\122\008\
    \118\006\145\006\206\006\175\004\018\006\033\006\147\006\148\006\
    \133\005\086\006\119\006\152\005\032\006\133\005\130\005\135\005\
    \035\006\116\006\019\006\034\006\154\005\153\005\133\005\142\004\
    \140\004\150\004\149\004\153\004\141\004\156\004\152\004\155\004\
    \225\004\083\005\157\004\138\004\145\004\151\004\144\004\143\004\
    \154\004\147\004\148\004\146\004\139\004\170\004\244\005\138\005\
    \009\006\245\005\137\005\163\004\178\004\178\004\243\005\172\004\
    \246\005\163\004\164\004\074\005\171\004\165\004\163\004\164\004\
    \169\004\169\004\165\004\173\004\169\004\174\004\177\004\035\005\
    \032\005\027\005\163\004\178\004\178\004\225\004\176\004\221\004\
    \163\004\216\004\190\004\209\004\191\004\163\004\193\004\169\004\
    \171\004\192\004\208\004\194\004\195\004\196\004\197\004\198\004\
    \198\004\200\004\201\004\202\004\203\004\204\004\205\004\206\004\
    \207\004\207\004\210\004\207\004\211\004\212\004\213\004\183\004\
    \182\004\186\004\214\004\214\004\185\004\189\004\198\004\198\004\
    \188\004\171\004\217\004\184\004\218\004\219\004\187\004\180\004\
    \181\004\179\004\220\004\207\004\222\004\223\004\224\004\162\004\
    \207\004\214\004\214\004\226\004\227\004\162\004\228\004\228\004\
    \017\005\013\005\162\004\009\005\000\005\249\004\240\004\236\004\
    \199\004\237\004\238\004\239\004\207\004\241\004\246\004\243\004\
    \244\004\242\004\245\004\207\004\247\004\228\004\228\004\248\004\
    \207\004\250\004\251\004\252\004\253\004\254\004\255\004\207\004\
    \001\005\002\005\003\005\004\005\005\005\006\005\215\004\007\005\
    \008\005\207\004\010\005\011\005\012\005\207\004\014\005\015\005\
    \016\005\207\004\018\005\019\005\020\005\021\005\022\005\023\005\
    \024\005\229\004\025\005\235\004\026\005\207\004\028\005\233\004\
    \030\005\207\004\029\005\031\005\207\004\231\004\158\004\033\005\
    \234\004\034\005\232\004\230\004\207\004\036\005\037\005\069\005\
    \042\005\039\005\040\005\041\005\207\004\043\005\044\005\038\005\
    \045\005\046\005\046\005\066\005\059\005\057\005\051\005\052\005\
    \053\005\054\005\055\005\056\005\207\004\058\005\207\004\060\005\
    \061\005\062\005\062\005\064\005\065\005\207\004\067\005\068\005\
    \046\005\046\005\061\005\070\005\071\005\071\005\073\005\207\004\
    \075\005\076\005\077\005\078\005\079\005\080\005\207\004\081\005\
    \062\005\062\005\082\005\207\004\084\005\085\005\114\005\088\005\
    \085\005\089\005\090\005\071\005\071\005\091\005\092\005\093\005\
    \094\005\095\005\095\005\113\005\106\005\099\005\100\005\101\005\
    \102\005\103\005\104\005\105\005\085\005\048\005\050\005\085\005\
    \207\004\107\005\108\005\047\005\049\005\109\005\110\005\111\005\
    \095\005\095\005\112\005\207\004\207\004\115\005\116\005\117\005\
    \118\005\119\005\120\005\123\005\063\005\124\005\121\005\121\005\
    \125\005\126\005\127\005\072\005\128\005\129\005\207\004\131\005\
    \132\005\084\005\207\004\136\005\133\005\151\005\139\005\140\005\
    \086\005\141\005\096\005\142\005\098\005\121\005\121\005\087\005\
    \143\005\143\005\190\004\145\005\146\005\147\005\148\005\149\005\
    \150\005\133\005\133\005\187\005\162\005\164\005\097\005\161\005\
    \155\005\156\005\157\005\158\005\192\004\159\005\160\005\143\005\
    \143\005\198\004\185\005\177\005\163\005\170\005\165\005\122\005\
    \166\005\171\005\167\005\168\005\169\005\133\005\175\005\172\005\
    \173\005\174\005\133\005\176\005\133\005\178\005\179\005\180\005\
    \198\004\181\005\181\005\183\005\184\005\133\005\186\005\188\005\
    \189\005\190\005\191\005\183\004\182\004\186\004\133\005\227\005\
    \185\004\144\005\199\005\197\005\188\004\194\005\198\005\184\004\
    \181\005\181\005\187\004\180\004\181\004\179\004\228\005\195\005\
    \210\005\196\005\133\005\217\004\200\005\193\005\209\005\201\005\
    \201\005\203\005\204\005\205\005\206\005\207\005\208\005\133\005\
    \212\005\211\005\133\005\213\005\214\005\215\005\216\005\217\005\
    \217\005\219\005\220\005\221\005\222\005\223\005\201\005\201\005\
    \224\005\225\005\226\005\133\005\231\005\229\005\133\005\230\005\
    \133\005\232\005\233\005\234\005\235\005\182\005\217\005\217\005\
    \235\005\237\005\238\005\239\005\240\005\241\005\242\005\186\005\
    \008\006\133\005\002\006\248\005\252\005\249\005\250\005\251\005\
    \133\005\253\005\254\005\235\005\255\005\000\006\001\006\235\005\
    \133\005\003\006\222\004\202\005\004\006\005\006\247\005\006\006\
    \007\006\133\005\133\005\010\006\218\005\011\006\012\006\013\006\
    \014\006\015\006\016\006\133\005\133\005\023\006\017\006\020\006\
    \021\006\022\006\207\004\029\005\024\006\025\006\026\006\027\006\
    \028\006\029\006\030\006\031\006\080\006\074\006\041\006\036\006\
    \081\006\040\006\037\006\133\005\070\006\236\005\062\006\054\006\
    \048\006\043\006\044\006\045\006\046\006\033\005\047\006\133\005\
    \061\006\049\006\050\006\051\006\052\006\053\006\133\005\055\006\
    \056\006\057\006\058\006\059\006\060\006\038\006\133\005\068\006\
    \063\006\064\006\065\006\066\006\067\006\133\005\039\006\069\006\
    \133\005\071\006\042\006\072\006\073\006\133\005\075\006\076\006\
    \077\006\078\006\079\006\133\005\083\006\082\006\133\005\084\006\
    \085\006\133\005\115\006\036\005\089\006\107\006\090\006\088\006\
    \091\006\092\006\093\006\094\006\094\006\038\005\101\006\097\006\
    \098\006\099\006\100\006\133\005\102\006\103\006\104\006\105\006\
    \106\006\133\005\108\006\110\006\112\006\111\006\052\005\109\006\
    \113\006\114\006\094\006\094\006\133\005\133\005\144\006\132\006\
    \128\006\120\006\121\006\122\006\123\006\124\006\125\006\126\006\
    \127\006\207\004\129\006\130\006\131\006\133\005\133\006\134\006\
    \135\006\138\006\136\006\136\006\139\006\140\006\141\006\142\006\
    \143\006\133\005\133\005\184\006\181\006\175\006\170\006\169\006\
    \151\006\156\006\152\006\153\006\095\006\154\006\155\006\048\005\
    \096\006\136\006\136\006\133\005\157\006\047\005\049\005\158\006\
    \160\006\164\006\161\006\084\005\162\006\163\006\133\005\165\006\
    \166\006\167\006\168\006\133\005\111\006\171\006\172\006\173\006\
    \174\006\159\006\133\005\176\006\177\006\178\006\179\006\180\006\
    \133\005\182\006\183\006\133\005\185\006\031\006\201\006\188\006\
    \189\006\190\006\190\006\192\006\193\006\194\006\195\006\196\006\
    \137\006\197\006\198\006\199\006\200\006\202\006\203\006\204\006\
    \205\006\186\005\133\005\112\008\209\006\111\008\211\006\241\006\
    \190\006\190\006\210\006\233\006\224\006\215\006\214\006\213\006\
    \216\006\216\006\221\006\212\006\219\006\220\006\225\006\200\006\
    \222\006\223\006\200\006\231\006\226\006\227\006\228\006\229\006\
    \230\006\133\005\232\006\133\005\234\006\235\006\236\006\216\006\
    \216\006\237\006\238\006\239\006\240\006\133\005\242\006\243\006\
    \244\006\109\008\105\008\101\008\084\008\200\006\200\006\093\008\
    \191\006\094\008\049\008\040\008\037\008\236\007\208\007\235\007\
    \161\007\067\007\016\008\048\008\050\008\017\008\192\007\244\006\
    \244\006\193\007\015\008\207\007\018\008\095\008\237\007\238\007\
    \107\007\162\007\046\007\031\006\035\007\023\007\160\007\106\007\
    \037\007\008\007\045\007\108\007\015\007\011\007\109\007\244\006\
    \010\007\012\007\036\007\013\007\014\007\031\006\044\007\016\007\
    \017\007\018\007\009\007\019\007\020\007\021\007\022\007\031\006\
    \024\007\007\007\025\007\026\007\029\007\030\007\027\007\031\007\
    \032\007\033\007\034\007\031\006\255\255\217\006\255\255\218\006\
    \248\006\252\006\002\007\255\006\253\006\004\007\000\007\249\006\
    \250\006\027\007\031\006\005\007\247\006\027\007\246\006\001\007\
    \245\006\003\007\254\006\006\007\041\007\251\006\040\007\038\007\
    \039\007\031\006\042\007\031\006\043\007\031\006\031\006\053\007\
    \027\007\048\007\047\007\031\006\049\007\050\007\051\007\052\007\
    \031\006\054\007\055\007\056\007\057\007\058\007\059\007\060\007\
    \061\007\061\007\063\007\064\007\065\007\066\007\031\006\102\007\
    \085\007\069\007\074\007\073\007\072\007\031\006\084\007\075\007\
    \076\007\076\007\078\007\070\007\079\007\080\007\103\007\061\007\
    \061\007\068\007\071\007\081\007\082\007\083\007\028\007\031\006\
    \087\007\086\007\031\006\088\007\089\007\090\007\091\007\076\007\
    \076\007\092\007\092\007\094\007\095\007\096\007\097\007\098\007\
    \099\007\100\007\101\007\031\006\031\006\104\007\105\007\031\006\
    \154\007\148\007\146\007\113\007\155\007\142\007\112\007\134\007\
    \092\007\092\007\126\007\120\007\115\007\116\007\117\007\118\007\
    \119\007\133\007\031\006\062\007\121\007\122\007\123\007\124\007\
    \125\007\031\006\127\007\128\007\077\007\129\007\130\007\255\255\
    \255\255\131\007\110\007\132\007\031\006\140\007\135\007\136\007\
    \137\007\138\007\139\007\111\007\031\006\141\007\093\007\114\007\
    \031\006\143\007\144\007\145\007\031\006\147\007\031\006\149\007\
    \150\007\151\007\152\007\153\007\031\006\157\007\156\007\031\006\
    \158\007\159\007\031\006\031\006\185\007\163\007\184\007\176\007\
    \166\007\167\007\174\007\164\007\165\007\169\007\170\007\171\007\
    \172\007\173\007\031\006\168\007\175\007\031\006\177\007\178\007\
    \179\007\180\007\181\007\182\007\183\007\031\006\031\006\186\007\
    \187\007\188\007\189\007\190\007\191\007\031\006\195\007\194\007\
    \031\006\196\007\197\007\198\007\199\007\199\007\201\007\202\007\
    \203\007\204\007\205\007\206\007\031\006\234\007\209\007\217\007\
    \211\007\210\007\212\007\213\007\214\007\215\007\216\007\031\006\
    \218\007\219\007\220\007\199\007\199\007\221\007\221\007\230\007\
    \224\007\225\007\226\007\227\007\228\007\229\007\031\006\231\007\
    \232\007\233\007\031\006\031\006\253\007\250\007\254\007\244\007\
    \239\007\240\007\241\007\242\007\221\007\221\007\243\007\031\006\
    \245\007\246\007\247\007\248\007\249\007\031\006\251\007\252\007\
    \031\006\002\008\255\007\000\008\001\008\200\007\031\006\003\008\
    \004\008\006\008\010\008\007\008\008\008\009\008\031\006\011\008\
    \012\008\013\008\014\008\031\006\036\008\031\006\030\008\020\008\
    \024\008\021\008\005\008\022\008\023\008\031\006\223\007\025\008\
    \026\008\027\008\222\007\028\008\029\008\031\006\031\008\032\008\
    \033\008\034\008\019\008\035\008\031\006\031\006\038\008\039\008\
    \031\006\041\008\042\008\043\008\044\008\046\008\031\006\047\008\
    \031\006\079\008\074\008\051\008\053\008\066\008\059\008\054\008\
    \045\008\055\008\060\008\056\008\057\008\058\008\031\006\064\008\
    \061\008\062\008\063\008\052\008\031\006\065\008\031\006\067\008\
    \068\008\069\008\070\008\070\008\072\008\073\008\031\006\075\008\
    \076\008\077\008\078\008\031\006\080\008\081\008\082\008\083\008\
    \031\006\085\008\086\008\087\008\088\008\089\008\090\008\091\008\
    \031\006\070\008\070\008\092\008\031\006\099\008\031\006\096\008\
    \097\008\098\008\031\006\100\008\031\006\102\008\103\008\104\008\
    \031\006\106\008\107\008\108\008\031\006\110\008\031\006\133\005\
    \113\008\114\008\115\008\116\008\117\008\118\008\133\005\121\008\
    \111\006\133\005\133\008\130\008\133\005\126\008\127\008\128\008\
    \129\008\111\006\133\005\132\008\111\006\125\008\131\008\134\008\
    \135\008\133\005\145\008\140\008\139\008\133\005\071\008\141\008\
    \142\008\143\008\144\008\133\005\146\008\147\008\148\008\149\008\
    \150\008\151\008\152\008\153\008\153\008\155\008\156\008\157\008\
    \158\008\133\005\160\008\161\008\133\005\163\008\164\008\165\008\
    \166\008\168\008\133\005\169\008\133\005\171\008\172\008\173\008\
    \133\005\195\008\153\008\153\008\167\008\183\008\180\008\179\008\
    \211\008\181\008\194\008\180\008\194\008\180\008\184\008\205\008\
    \194\008\187\008\204\008\186\008\180\008\194\008\180\008\184\008\
    \187\008\194\008\187\008\207\008\183\008\209\008\186\008\180\008\
    \255\255\194\008\184\008\194\008\231\008\231\008\182\008\187\008\
    \187\008\179\008\186\008\193\008\194\008\187\008\179\008\187\008\
    \179\008\187\008\188\008\254\008\185\008\186\008\154\008\179\008\
    \178\008\179\008\195\008\188\008\187\008\255\255\187\008\185\008\
    \191\008\214\008\179\008\204\008\187\008\194\008\194\008\214\008\
    \189\008\194\008\255\255\191\008\203\008\198\008\190\008\207\008\
    \199\008\195\008\208\008\187\008\209\008\203\008\198\008\210\008\
    \214\008\199\008\204\008\176\008\194\008\191\008\214\008\214\008\
    \253\008\255\008\252\008\203\008\231\008\213\008\243\008\244\008\
    \231\008\179\008\242\008\213\008\203\008\237\008\239\008\231\008\
    \202\008\235\008\239\008\234\008\250\008\240\008\214\008\251\008\
    \232\008\202\008\255\008\229\008\243\008\244\008\230\008\201\008\
    \241\008\005\009\006\009\213\008\231\008\011\009\012\009\003\009\
    \021\009\020\009\009\009\019\009\038\009\037\009\036\009\031\009\
    \030\009\255\008\031\009\220\008\222\008\224\008\223\008\032\009\
    \216\008\226\008\231\008\225\008\231\008\231\008\200\008\231\008\
    \206\008\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\033\009\237\008\221\008\042\009\238\008\
    \054\009\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\228\008\231\008\216\008\041\009\
    \218\008\045\009\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\216\008\216\008\216\008\216\008\
    \216\008\216\008\216\008\216\008\227\008\248\008\246\008\249\008\
    \249\008\247\008\061\009\249\008\249\008\246\008\248\008\246\008\
    \249\008\060\009\247\008\054\009\177\008\236\008\054\009\097\009\
    \096\009\249\008\249\008\098\009\248\008\249\008\249\008\103\009\
    \088\009\123\009\122\009\249\008\245\008\248\008\245\008\121\009\
    \118\009\028\009\114\009\245\008\029\009\245\008\255\255\105\009\
    \249\008\035\009\033\009\035\009\033\009\034\009\113\009\034\009\
    \245\008\255\255\048\009\048\009\018\009\115\009\048\009\017\009\
    \119\009\079\009\080\009\015\009\078\009\077\009\105\009\026\009\
    \035\009\025\009\035\009\255\255\119\009\132\009\129\009\120\009\
    \255\255\048\009\090\009\090\009\070\009\071\009\255\255\069\009\
    \068\009\255\255\088\009\255\255\143\009\027\009\079\009\080\009\
    \135\009\078\009\077\009\140\009\255\255\087\009\144\009\159\009\
    \186\009\090\009\255\255\174\009\255\255\149\009\255\255\165\009\
    \095\009\255\255\094\009\126\009\171\009\004\009\169\009\135\009\
    \010\009\128\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\149\009\172\009\175\009\182\009\
    \179\009\217\008\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\183\009\255\255\
    \184\009\255\255\187\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\100\009\098\009\
    \224\009\225\009\099\009\223\009\222\009\159\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \201\009\202\009\208\009\200\009\199\009\100\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\208\009\255\255\233\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\016\009\048\009\048\009\237\009\168\009\048\009\024\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \242\009\048\009\220\009\221\009\168\009\219\009\218\009\255\255\
    \255\255\057\009\244\009\056\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \093\009\090\009\255\255\127\009\216\009\217\009\245\009\215\009\
    \214\009\240\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\058\009\248\009\059\009\249\009\
    \054\009\167\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\250\009\252\009\
    \241\009\002\010\000\010\000\010\012\010\003\010\255\255\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\007\010\007\010\
    \013\010\014\010\041\010\053\009\052\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\051\009\
    \050\009\042\010\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\049\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\055\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\241\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\040\010\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\100\009\098\009\001\010\057\010\099\009\
    \057\010\062\009\072\010\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\232\009\236\009\072\010\
    \000\000\000\000\100\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\014\010\000\000\
    \000\000\015\010\054\009\000\000\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \000\000\000\000\000\000\000\000\054\009\000\000\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\000\000\067\009\066\009\088\010\
    \089\010\000\000\087\010\086\010\117\009\115\009\000\000\000\000\
    \116\009\065\009\064\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\117\009\000\000\063\009\117\009\115\009\
    \000\000\000\000\116\009\131\009\129\009\131\009\129\009\130\009\
    \000\000\130\009\146\009\144\009\000\000\000\000\145\009\142\009\
    \140\009\255\255\000\000\141\009\000\000\117\009\000\000\076\009\
    \075\009\000\000\131\009\000\010\131\009\000\000\142\009\140\009\
    \000\000\146\009\141\009\074\009\073\009\000\000\142\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\255\255\255\255\142\009\000\000\072\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \000\000\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\000\000\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\000\000\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\000\000\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\000\000\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\000\000\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\086\009\086\009\000\000\149\009\
    \086\009\086\009\086\009\000\000\242\009\086\009\255\255\255\255\
    \105\009\103\009\017\010\091\010\104\009\019\010\135\009\000\000\
    \000\000\246\009\000\000\086\009\017\010\085\009\149\009\019\010\
    \086\009\023\010\255\255\242\009\025\010\255\255\246\009\105\009\
    \045\010\083\009\107\009\109\009\108\009\135\009\087\009\112\009\
    \246\009\111\009\146\009\144\009\000\000\137\009\145\009\000\000\
    \168\009\169\009\250\009\139\009\170\009\246\009\000\000\045\010\
    \000\000\171\009\172\009\174\009\175\009\173\009\000\000\176\009\
    \043\010\146\009\183\009\184\009\188\009\188\009\185\009\168\009\
    \188\009\250\009\000\000\168\009\169\009\152\009\165\009\170\009\
    \171\009\166\009\174\009\251\009\252\009\241\009\000\000\043\010\
    \253\009\183\009\110\009\188\009\000\000\000\000\000\000\000\000\
    \243\009\042\010\168\009\000\000\152\009\000\000\153\009\045\010\
    \054\010\011\010\007\010\252\009\161\009\010\010\000\000\000\000\
    \000\000\000\000\162\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\247\009\045\010\054\010\
    \011\010\000\000\136\009\000\000\167\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\167\009\
    \000\000\164\009\000\000\159\009\091\010\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\160\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\091\010\000\000\
    \000\000\181\009\179\009\082\009\229\009\180\009\255\255\000\000\
    \255\255\023\010\011\010\007\010\025\010\255\255\010\010\106\009\
    \255\255\255\255\000\000\000\000\000\000\138\009\158\009\157\009\
    \181\009\000\000\255\255\229\009\000\000\228\009\000\000\255\255\
    \000\000\011\010\156\009\155\009\000\000\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\154\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\163\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\000\000\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\000\000\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\065\010\066\010\
    \000\000\064\010\063\010\000\000\189\009\000\000\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \084\010\085\010\255\255\083\010\082\010\000\000\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\000\000\000\000\000\000\000\000\159\009\000\000\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\000\000\181\009\179\009\080\010\081\010\180\009\079\010\
    \078\010\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \178\009\000\000\000\000\255\255\000\000\000\000\000\000\255\255\
    \000\000\181\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\000\000\177\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \194\009\193\009\000\000\000\000\000\000\000\000\034\010\000\000\
    \159\009\033\010\000\000\000\000\192\009\191\009\000\000\189\009\
    \000\000\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\000\000\000\000\000\000\000\000\000\000\
    \190\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\000\000\000\000\000\000\000\000\
    \159\009\000\000\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\195\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\000\000\000\000\
    \000\000\000\000\208\009\000\000\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \000\000\000\000\000\000\194\009\193\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\192\009\
    \191\009\178\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\000\000\190\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\000\000\207\009\206\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\205\009\204\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\000\000\255\255\203\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\000\000\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\000\000\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\000\000\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\000\000\000\000\000\000\
    \000\000\000\000\000\000\189\009\000\000\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \000\000\000\000\000\000\000\000\159\009\000\000\159\009\159\009\
    \159\009\159\009\196\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\194\009\
    \193\009\000\000\000\000\000\000\000\000\000\000\000\000\159\009\
    \000\000\000\000\000\000\192\009\191\009\000\000\189\009\000\000\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\000\000\000\000\000\000\000\000\000\000\190\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\029\010\000\000\000\000\033\010\159\009\
    \000\000\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\197\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\009\010\007\010\034\010\017\010\008\010\033\010\
    \019\010\043\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\000\000\000\000\
    \000\000\009\010\194\009\193\009\000\000\000\000\000\000\000\000\
    \043\010\000\000\159\009\000\000\000\000\000\000\192\009\191\009\
    \021\010\189\009\044\010\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\000\000\000\000\000\000\
    \000\000\000\000\190\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\006\010\035\010\
    \000\000\000\000\159\009\000\000\159\009\159\009\159\009\159\009\
    \198\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\000\000\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\000\000\000\000\029\010\
    \017\010\255\255\033\010\019\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\000\000\194\009\193\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\159\009\000\000\000\000\
    \000\000\192\009\191\009\022\010\189\009\000\000\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \000\000\000\000\000\000\000\000\000\000\190\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\255\255\255\255\255\255\000\000\159\009\000\000\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \000\000\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \000\000\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \194\009\193\009\034\010\000\000\000\000\033\010\000\000\000\000\
    \208\009\000\000\000\000\000\000\192\009\191\009\000\000\189\009\
    \000\000\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\000\000\000\000\000\000\000\000\000\000\
    \190\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\000\000\255\255\255\255\000\000\
    \208\009\000\000\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\000\000\000\000\213\009\212\009\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\211\009\
    \210\009\000\000\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\209\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \255\255\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \000\000\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\016\010\017\010\023\010\055\010\019\010\025\010\031\010\
    \029\010\031\010\029\010\030\010\034\010\032\010\029\010\033\010\
    \029\010\037\010\000\000\033\010\000\000\000\000\000\000\000\000\
    \016\010\000\000\000\000\055\010\000\000\000\000\031\010\000\000\
    \031\010\000\000\000\000\000\000\000\000\020\010\022\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\000\000\000\000\054\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\000\000\028\010\000\000\028\010\053\010\036\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\052\010\051\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\050\010\049\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\000\000\
    \000\000\048\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\000\000\255\255\000\000\255\255\000\000\255\255\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \000\000\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\000\000\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\055\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\000\000\000\000\055\010\000\000\000\000\000\000\
    \000\000\000\000\000\000\057\010\000\000\000\000\000\000\000\000\
    \000\000\000\000\056\010\000\000\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\054\010\000\000\
    \000\000\000\000\000\000\000\000\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\000\000\
    \000\000\000\000\000\000\057\010\000\000\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\000\000\000\000\000\000\000\000\072\010\000\000\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\062\010\061\010\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\060\010\059\010\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\058\010\000\000\
    \000\000\000\000\000\000\055\010\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \071\010\070\010\055\010\000\000\000\000\000\000\000\000\000\000\
    \000\000\057\010\000\000\000\000\069\010\068\010\000\000\000\000\
    \056\010\000\000\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\054\010\000\000\000\000\000\000\
    \067\010\000\000\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\000\000\000\000\000\000\
    \000\000\057\010\000\000\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\062\010\061\010\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \060\010\059\010\000\000\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\058\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\000\000\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\000\000\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\000\000\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\000\000\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\055\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\055\010\000\000\
    \000\000\000\000\000\000\000\000\000\000\072\010\000\000\000\000\
    \000\000\000\000\000\000\000\000\056\010\000\000\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \054\010\000\000\000\000\000\000\000\000\000\000\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\000\000\000\000\000\000\000\000\072\010\000\000\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \077\010\076\010\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\075\010\074\010\000\000\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \073\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\000\000\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\000\000\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\024\000\025\000\000\000\024\000\048\000\
    \054\000\054\000\066\000\105\000\054\000\067\000\068\000\069\000\
    \070\000\075\000\084\000\085\000\086\000\112\000\087\000\088\000\
    \000\000\095\000\103\000\103\000\114\000\123\000\103\000\054\000\
    \000\000\066\000\023\000\047\000\067\000\068\000\069\000\070\000\
    \075\000\084\000\085\000\086\000\090\000\087\000\088\000\102\000\
    \095\000\103\000\109\000\114\000\085\000\086\000\087\000\087\000\
    \088\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\120\000\116\002\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\002\
    \202\002\145\003\204\003\159\004\116\002\161\004\162\004\165\004\
    \167\004\004\000\006\000\009\000\006\000\007\000\010\000\008\000\
    \012\000\018\000\012\000\028\000\012\000\042\000\010\000\003\000\
    \005\000\008\000\011\000\014\000\009\000\007\000\012\000\013\000\
    \006\000\009\000\010\000\012\000\012\000\015\000\013\000\017\000\
    \014\000\016\000\017\000\019\000\027\000\029\000\035\000\017\000\
    \013\000\017\000\021\000\015\000\019\000\019\000\031\000\013\000\
    \020\000\016\000\051\000\020\000\021\000\022\000\026\000\026\000\
    \022\000\021\000\026\000\030\000\022\000\030\000\031\000\032\000\
    \022\000\033\000\034\000\022\000\044\000\052\000\038\000\044\000\
    \040\000\032\000\034\000\039\000\036\000\026\000\036\000\057\000\
    \036\000\030\000\033\000\038\000\037\000\026\000\034\000\033\000\
    \040\000\039\000\036\000\037\000\058\000\060\000\041\000\036\000\
    \036\000\041\000\061\000\062\000\063\000\037\000\041\000\064\000\
    \041\000\065\000\071\000\072\000\037\000\073\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\074\000\
    \077\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\043\000\045\000\049\000\049\000\
    \050\000\078\000\049\000\056\000\053\000\043\000\043\000\045\000\
    \046\000\059\000\079\000\046\000\045\000\080\000\081\000\046\000\
    \076\000\082\000\083\000\046\000\092\000\049\000\046\000\050\000\
    \000\000\093\000\056\000\053\000\094\000\049\000\089\000\097\000\
    \059\000\050\000\098\000\099\000\067\000\068\000\069\000\076\000\
    \100\000\096\000\085\000\086\000\106\000\087\000\107\000\113\000\
    \116\000\117\000\118\000\126\000\129\000\089\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\089\000\
    \096\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\101\000\108\000\110\000\110\000\
    \115\000\124\000\110\000\127\000\119\000\124\000\131\000\132\000\
    \133\000\050\000\121\000\121\000\050\000\053\000\121\000\128\000\
    \053\000\127\000\128\000\101\000\108\000\110\000\134\000\115\000\
    \135\000\130\000\056\000\119\000\056\000\101\000\108\000\136\000\
    \059\000\121\000\059\000\130\000\137\000\119\000\138\000\076\000\
    \139\000\139\000\141\000\142\000\139\000\143\000\128\000\144\000\
    \145\000\147\000\148\000\149\000\150\000\151\000\128\000\128\000\
    \096\000\152\000\153\000\128\000\154\000\155\000\156\000\139\000\
    \157\000\158\000\159\000\096\000\160\000\161\000\162\000\163\000\
    \164\000\165\000\166\000\167\000\168\000\169\000\170\000\171\000\
    \172\000\173\000\174\000\175\000\177\000\178\000\179\000\180\000\
    \181\000\182\000\183\000\184\000\185\000\186\000\187\000\188\000\
    \189\000\190\000\191\000\192\000\193\000\192\000\194\000\195\000\
    \190\000\196\000\115\000\197\000\198\000\199\000\200\000\195\000\
    \201\000\202\000\203\000\204\000\205\000\206\000\207\000\208\000\
    \209\000\210\000\211\000\207\000\210\000\212\000\213\000\214\000\
    \215\000\216\000\197\000\198\000\217\000\218\000\219\000\220\000\
    \221\000\222\000\223\000\224\000\225\000\226\000\227\000\228\000\
    \227\000\229\000\230\000\231\000\233\000\234\000\232\000\235\000\
    \216\000\236\000\237\000\217\000\238\000\239\000\240\000\241\000\
    \242\000\243\000\244\000\245\000\246\000\247\000\198\000\198\000\
    \198\000\248\000\231\000\198\000\198\000\232\000\249\000\198\000\
    \250\000\251\000\198\000\252\000\253\000\198\000\198\000\198\000\
    \198\000\254\000\245\000\246\000\217\000\255\000\000\001\001\001\
    \002\001\003\001\004\001\005\001\002\001\089\000\006\001\007\001\
    \008\001\009\001\010\001\011\001\012\001\013\001\014\001\015\001\
    \016\001\017\001\018\001\019\001\020\001\021\001\022\001\023\001\
    \024\001\232\000\025\001\026\001\027\001\028\001\246\000\029\001\
    \246\000\030\001\031\001\032\001\246\000\033\001\034\001\035\001\
    \036\001\037\001\246\000\038\001\039\001\246\000\040\001\246\000\
    \246\000\041\001\042\001\043\001\044\001\045\001\046\001\047\001\
    \048\001\045\001\049\001\050\001\051\001\051\001\052\001\053\001\
    \054\001\055\001\056\001\057\001\058\001\059\001\051\001\060\001\
    \061\001\062\001\063\001\064\001\065\001\066\001\067\001\068\001\
    \069\001\070\001\071\001\072\001\073\001\074\001\075\001\076\001\
    \077\001\078\001\079\001\080\001\081\001\082\001\083\001\061\001\
    \062\001\084\001\085\001\086\001\087\001\088\001\089\001\090\001\
    \091\001\092\001\093\001\094\001\095\001\088\001\096\001\077\001\
    \078\001\097\001\098\001\099\001\100\001\102\001\103\001\101\001\
    \104\001\105\001\086\001\087\001\106\001\107\001\108\001\109\001\
    \110\001\111\001\112\001\113\001\114\001\115\001\116\001\117\001\
    \118\001\119\001\120\001\100\001\062\001\062\001\101\001\121\001\
    \122\001\123\001\062\001\062\001\124\001\125\001\126\001\110\001\
    \111\001\127\001\128\001\129\001\130\001\131\001\132\001\133\001\
    \134\001\135\001\138\001\078\001\139\001\136\001\137\001\140\001\
    \141\001\142\001\087\001\143\001\144\001\145\001\146\001\147\001\
    \148\001\149\001\150\001\151\001\152\001\153\001\157\001\101\001\
    \154\001\111\001\155\001\111\001\136\001\137\001\101\001\156\001\
    \156\001\158\001\159\001\160\001\161\001\162\001\163\001\164\001\
    \165\001\165\001\166\001\165\001\168\001\111\001\169\001\154\001\
    \167\001\155\001\170\001\167\001\171\001\172\001\173\001\174\001\
    \165\001\175\001\176\001\177\001\178\001\175\001\137\001\179\001\
    \180\001\181\001\182\001\183\001\184\001\185\001\186\001\187\001\
    \188\001\189\001\190\001\191\001\172\001\194\001\195\001\196\001\
    \192\001\193\001\197\001\199\001\155\001\155\001\155\001\200\001\
    \201\001\155\001\155\001\202\001\203\001\155\001\207\001\204\001\
    \155\001\205\001\208\001\155\001\155\001\155\001\155\001\192\001\
    \193\001\204\001\209\001\206\001\207\001\204\001\210\001\204\001\
    \205\001\206\001\211\001\212\001\213\001\214\001\215\001\216\001\
    \217\001\218\001\219\001\220\001\221\001\222\001\223\001\224\001\
    \225\001\226\001\227\001\228\001\229\001\230\001\231\001\232\001\
    \233\001\234\001\212\001\213\001\235\001\236\001\237\001\238\001\
    \239\001\240\001\239\001\241\001\242\001\243\001\244\001\245\001\
    \246\001\247\001\228\001\229\001\193\001\248\001\249\001\250\001\
    \251\001\252\001\253\001\254\001\255\001\000\002\001\002\002\002\
    \003\002\004\002\005\002\006\002\007\002\008\002\009\002\246\001\
    \247\001\010\002\011\002\012\002\013\002\014\002\002\002\213\001\
    \015\002\016\002\002\002\017\002\018\002\019\002\020\002\021\002\
    \229\001\022\002\023\002\024\002\025\002\026\002\027\002\028\002\
    \029\002\030\002\027\002\031\002\032\002\033\002\034\002\031\002\
    \035\002\036\002\037\002\038\002\039\002\040\002\041\002\042\002\
    \044\002\045\002\045\002\046\002\047\002\045\002\247\001\048\002\
    \049\002\050\002\051\002\045\002\052\002\053\002\054\002\055\002\
    \056\002\057\002\058\002\059\002\060\002\061\002\062\002\063\002\
    \064\002\065\002\066\002\067\002\068\002\055\002\066\002\070\002\
    \071\002\051\002\072\002\052\002\073\002\074\002\075\002\076\002\
    \077\002\078\002\079\002\080\002\081\002\082\002\083\002\084\002\
    \085\002\086\002\087\002\088\002\089\002\090\002\091\002\092\002\
    \095\002\096\002\093\002\094\002\097\002\098\002\099\002\100\002\
    \101\002\102\002\103\002\104\002\105\002\106\002\107\002\108\002\
    \109\002\108\002\110\002\111\002\052\002\112\002\113\002\052\002\
    \052\002\093\002\094\002\114\002\115\002\052\002\052\002\117\002\
    \118\002\119\002\108\002\120\002\121\002\122\002\123\002\124\002\
    \125\002\126\002\127\002\129\002\130\002\131\002\132\002\133\002\
    \134\002\135\002\136\002\137\002\138\002\139\002\140\002\141\002\
    \142\002\143\002\144\002\145\002\146\002\147\002\148\002\149\002\
    \150\002\151\002\152\002\153\002\154\002\155\002\156\002\157\002\
    \094\002\158\002\159\002\160\002\161\002\162\002\163\002\164\002\
    \165\002\166\002\167\002\168\002\169\002\170\002\171\002\172\002\
    \173\002\175\002\174\002\176\002\177\002\172\002\178\002\179\002\
    \180\002\181\002\174\002\174\002\182\002\163\002\164\002\174\002\
    \178\002\183\002\184\002\185\002\186\002\187\002\188\002\190\002\
    \191\002\192\002\193\002\194\002\195\002\196\002\197\002\198\002\
    \199\002\200\002\201\002\203\002\204\002\205\002\206\002\207\002\
    \208\002\209\002\210\002\211\002\212\002\188\002\190\002\213\002\
    \214\002\213\002\216\002\217\002\215\002\220\002\219\002\221\002\
    \219\002\222\002\221\002\188\002\190\002\215\002\215\002\208\002\
    \209\002\164\002\220\002\218\002\224\002\213\002\218\002\219\002\
    \219\002\223\002\222\002\218\002\223\002\218\002\226\002\222\002\
    \223\002\228\002\229\002\230\002\223\002\231\002\225\002\223\002\
    \226\002\232\002\233\002\234\002\235\002\236\002\225\002\237\002\
    \238\002\209\002\209\002\209\002\209\002\209\002\209\002\209\002\
    \209\002\209\002\225\002\239\002\209\002\209\002\227\002\209\002\
    \209\002\209\002\209\002\209\002\209\002\227\002\209\002\240\002\
    \241\002\242\002\243\002\244\002\245\002\246\002\247\002\227\002\
    \248\002\249\002\188\002\190\002\188\002\190\002\227\002\250\002\
    \251\002\252\002\253\002\254\002\255\002\000\003\001\003\002\003\
    \001\003\003\003\004\003\005\003\006\003\247\002\002\003\248\002\
    \007\003\008\003\009\003\010\003\011\003\012\003\013\003\014\003\
    \015\003\016\003\017\003\018\003\019\003\020\003\021\003\022\003\
    \023\003\024\003\025\003\026\003\027\003\028\003\029\003\030\003\
    \031\003\033\003\034\003\032\003\036\003\037\003\035\003\038\003\
    \034\003\039\003\040\003\041\003\042\003\032\003\043\003\044\003\
    \033\003\025\003\026\003\032\003\035\003\248\002\045\003\046\003\
    \047\003\048\003\049\003\050\003\051\003\052\003\053\003\054\003\
    \055\003\040\003\041\003\056\003\057\003\058\003\059\003\060\003\
    \061\003\062\003\063\003\064\003\065\003\066\003\067\003\068\003\
    \069\003\070\003\071\003\072\003\073\003\075\003\071\003\076\003\
    \077\003\074\003\056\003\057\003\074\003\078\003\079\003\080\003\
    \081\003\076\003\082\003\083\003\084\003\026\003\085\003\086\003\
    \087\003\088\003\089\003\090\003\091\003\092\003\041\003\093\003\
    \094\003\095\003\096\003\097\003\098\003\099\003\100\003\101\003\
    \074\003\102\003\103\003\104\003\105\003\188\002\190\002\106\003\
    \057\003\074\003\107\003\108\003\109\003\074\003\110\003\111\003\
    \112\003\113\003\114\003\115\003\116\003\117\003\118\003\119\003\
    \120\003\121\003\122\003\123\003\124\003\125\003\126\003\127\003\
    \128\003\130\003\129\003\131\003\132\003\127\003\129\003\133\003\
    \134\003\135\003\136\003\137\003\138\003\131\003\139\003\140\003\
    \141\003\142\003\143\003\144\003\146\003\147\003\148\003\149\003\
    \150\003\151\003\152\003\153\003\154\003\155\003\156\003\157\003\
    \158\003\159\003\160\003\161\003\162\003\163\003\164\003\165\003\
    \166\003\167\003\168\003\169\003\170\003\171\003\172\003\173\003\
    \174\003\175\003\173\003\176\003\177\003\178\003\179\003\180\003\
    \181\003\182\003\183\003\184\003\163\003\164\003\185\003\186\003\
    \187\003\188\003\189\003\190\003\191\003\192\003\193\003\194\003\
    \195\003\196\003\197\003\198\003\199\003\200\003\201\003\200\003\
    \202\003\203\003\205\003\206\003\207\003\185\003\186\003\208\003\
    \209\003\210\003\211\003\212\003\213\003\214\003\215\003\216\003\
    \217\003\218\003\219\003\220\003\221\003\222\003\164\003\223\003\
    \224\003\225\003\226\003\227\003\228\003\229\003\230\003\231\003\
    \232\003\233\003\234\003\235\003\236\003\237\003\238\003\239\003\
    \240\003\241\003\225\003\242\003\243\003\244\003\245\003\186\003\
    \246\003\247\003\248\003\186\003\249\003\250\003\251\003\252\003\
    \253\003\254\003\239\003\255\003\000\004\001\004\002\004\003\004\
    \004\004\005\004\006\004\007\004\008\004\009\004\010\004\011\004\
    \012\004\013\004\014\004\015\004\015\004\016\004\017\004\018\004\
    \008\004\019\004\017\004\020\004\021\004\022\004\023\004\024\004\
    \025\004\026\004\027\004\015\004\028\004\029\004\030\004\031\004\
    \032\004\033\004\034\004\035\004\036\004\037\004\038\004\039\004\
    \040\004\041\004\042\004\043\004\044\004\045\004\046\004\047\004\
    \048\004\049\004\050\004\051\004\052\004\053\004\054\004\055\004\
    \056\004\034\004\035\004\055\004\057\004\058\004\059\004\060\004\
    \061\004\062\004\063\004\064\004\065\004\066\004\067\004\068\004\
    \069\004\070\004\071\004\072\004\073\004\074\004\075\004\076\004\
    \077\004\078\004\079\004\080\004\081\004\082\004\083\004\084\004\
    \085\004\086\004\087\004\088\004\089\004\090\004\091\004\092\004\
    \093\004\094\004\095\004\096\004\097\004\098\004\099\004\100\004\
    \101\004\102\004\103\004\104\004\105\004\106\004\035\004\107\004\
    \108\004\087\004\088\004\109\004\110\004\111\004\112\004\113\004\
    \114\004\108\004\115\004\114\004\116\004\117\004\118\004\119\004\
    \120\004\121\004\122\004\123\004\124\004\125\004\126\004\127\004\
    \128\004\129\004\130\004\131\004\132\004\133\004\134\004\135\004\
    \135\004\137\004\126\004\135\004\138\004\139\004\140\004\141\004\
    \145\004\142\004\143\004\142\004\144\004\146\004\154\004\141\004\
    \171\004\088\004\145\004\147\004\173\004\147\004\135\004\147\004\
    \137\004\148\004\143\004\141\004\149\004\144\004\135\004\142\004\
    \148\004\147\004\144\004\174\004\151\004\150\004\147\004\147\004\
    \150\004\149\004\148\004\155\004\150\004\176\004\179\004\170\004\
    \150\004\148\004\151\004\150\004\155\004\155\004\170\004\135\004\
    \135\004\135\004\135\004\135\004\135\004\135\004\135\004\135\004\
    \152\004\180\004\135\004\135\004\135\004\135\004\135\004\135\004\
    \135\004\135\004\135\004\135\004\135\004\157\004\153\004\156\004\
    \152\004\153\004\156\004\163\004\177\004\178\004\153\004\157\004\
    \153\004\160\004\160\004\181\004\157\004\160\004\164\004\164\004\
    \168\004\168\004\164\004\172\004\168\004\172\004\175\004\182\004\
    \183\004\184\004\163\004\177\004\178\004\185\004\175\004\186\004\
    \160\004\187\004\189\004\188\004\190\004\164\004\191\004\168\004\
    \188\004\191\004\192\004\193\004\194\004\195\004\196\004\197\004\
    \198\004\199\004\200\004\201\004\202\004\203\004\204\004\205\004\
    \206\004\208\004\209\004\210\004\209\004\211\004\212\004\178\004\
    \178\004\178\004\213\004\214\004\178\004\178\004\197\004\198\004\
    \178\004\215\004\216\004\178\004\217\004\218\004\178\004\178\004\
    \178\004\178\004\219\004\220\004\221\004\222\004\223\004\163\004\
    \224\004\213\004\214\004\225\004\226\004\160\004\227\004\228\004\
    \229\004\230\004\164\004\231\004\232\004\233\004\234\004\235\004\
    \198\004\236\004\237\004\238\004\239\004\240\004\241\004\242\004\
    \243\004\240\004\244\004\245\004\246\004\227\004\228\004\247\004\
    \248\004\249\004\250\004\251\004\252\004\253\004\254\004\255\004\
    \000\005\001\005\002\005\003\005\004\005\005\005\214\004\006\005\
    \007\005\008\005\009\005\010\005\011\005\012\005\013\005\014\005\
    \015\005\016\005\017\005\018\005\019\005\020\005\021\005\022\005\
    \023\005\228\004\024\005\228\004\025\005\026\005\027\005\228\004\
    \028\005\029\005\027\005\030\005\031\005\228\004\135\004\032\005\
    \228\004\033\005\228\004\228\004\034\005\035\005\035\005\036\005\
    \037\005\038\005\039\005\040\005\041\005\042\005\043\005\035\005\
    \044\005\045\005\046\005\047\005\048\005\049\005\050\005\051\005\
    \052\005\053\005\054\005\055\005\056\005\057\005\058\005\059\005\
    \060\005\061\005\062\005\063\005\064\005\065\005\066\005\067\005\
    \045\005\046\005\068\005\069\005\070\005\071\005\072\005\073\005\
    \074\005\075\005\076\005\077\005\078\005\079\005\072\005\080\005\
    \061\005\062\005\081\005\082\005\083\005\084\005\086\005\087\005\
    \085\005\088\005\089\005\070\005\071\005\090\005\091\005\092\005\
    \093\005\094\005\095\005\096\005\097\005\098\005\099\005\100\005\
    \101\005\102\005\103\005\104\005\084\005\046\005\046\005\085\005\
    \105\005\106\005\107\005\046\005\046\005\108\005\109\005\110\005\
    \094\005\095\005\111\005\112\005\113\005\114\005\115\005\116\005\
    \117\005\118\005\119\005\122\005\062\005\123\005\120\005\121\005\
    \124\005\125\005\126\005\071\005\127\005\128\005\129\005\130\005\
    \131\005\132\005\134\005\135\005\136\005\137\005\138\005\139\005\
    \085\005\140\005\095\005\141\005\095\005\120\005\121\005\085\005\
    \142\005\143\005\144\005\144\005\145\005\146\005\147\005\148\005\
    \149\005\150\005\151\005\152\005\153\005\153\005\095\005\153\005\
    \154\005\155\005\156\005\157\005\155\005\158\005\159\005\142\005\
    \143\005\160\005\161\005\162\005\153\005\163\005\164\005\121\005\
    \165\005\163\005\166\005\167\005\168\005\169\005\170\005\171\005\
    \172\005\173\005\174\005\175\005\176\005\177\005\178\005\179\005\
    \160\005\180\005\181\005\182\005\183\005\184\005\185\005\187\005\
    \188\005\189\005\190\005\143\005\143\005\143\005\191\005\193\005\
    \143\005\143\005\196\005\195\005\143\005\192\005\197\005\143\005\
    \180\005\181\005\143\005\143\005\143\005\143\005\193\005\192\005\
    \194\005\195\005\198\005\192\005\199\005\192\005\194\005\200\005\
    \201\005\202\005\203\005\204\005\205\005\206\005\207\005\208\005\
    \209\005\210\005\211\005\212\005\213\005\214\005\215\005\216\005\
    \217\005\218\005\219\005\220\005\221\005\222\005\200\005\201\005\
    \223\005\224\005\225\005\226\005\227\005\228\005\227\005\229\005\
    \230\005\231\005\232\005\233\005\234\005\181\005\216\005\217\005\
    \235\005\236\005\237\005\238\005\239\005\240\005\241\005\242\005\
    \243\005\244\005\245\005\246\005\247\005\248\005\249\005\250\005\
    \251\005\252\005\253\005\234\005\254\005\255\005\000\006\235\005\
    \001\006\002\006\246\005\201\005\003\006\004\006\246\005\005\006\
    \006\006\007\006\008\006\009\006\217\005\010\006\011\006\012\006\
    \013\006\014\006\015\006\016\006\017\006\018\006\015\006\019\006\
    \020\006\021\006\022\006\019\006\023\006\024\006\025\006\026\006\
    \027\006\028\006\029\006\030\006\032\006\033\006\034\006\035\006\
    \032\006\034\006\036\006\037\006\038\006\235\005\039\006\040\006\
    \041\006\042\006\043\006\044\006\045\006\035\006\046\006\047\006\
    \039\006\048\006\049\006\050\006\051\006\052\006\053\006\054\006\
    \055\006\056\006\057\006\058\006\059\006\034\006\060\006\061\006\
    \062\006\063\006\064\006\065\006\066\006\067\006\034\006\068\006\
    \069\006\070\006\034\006\071\006\072\006\073\006\074\006\075\006\
    \076\006\077\006\078\006\079\006\080\006\081\006\082\006\083\006\
    \084\006\085\006\086\006\087\006\087\006\088\006\089\006\087\006\
    \090\006\091\006\092\006\093\006\094\006\087\006\095\006\096\006\
    \097\006\098\006\099\006\100\006\101\006\102\006\103\006\104\006\
    \105\006\106\006\107\006\108\006\109\006\110\006\097\006\108\006\
    \112\006\113\006\093\006\094\006\114\006\115\006\116\006\117\006\
    \118\006\119\006\120\006\121\006\122\006\123\006\124\006\125\006\
    \126\006\127\006\128\006\129\006\130\006\131\006\132\006\133\006\
    \134\006\137\006\135\006\136\006\138\006\139\006\140\006\141\006\
    \142\006\143\006\144\006\145\006\146\006\147\006\148\006\149\006\
    \150\006\151\006\150\006\152\006\094\006\153\006\154\006\094\006\
    \094\006\135\006\136\006\155\006\156\006\094\006\094\006\157\006\
    \158\006\159\006\160\006\150\006\161\006\162\006\163\006\164\006\
    \165\006\166\006\167\006\168\006\169\006\170\006\171\006\172\006\
    \173\006\158\006\174\006\175\006\176\006\177\006\178\006\179\006\
    \180\006\181\006\182\006\183\006\184\006\185\006\186\006\187\006\
    \188\006\189\006\190\006\191\006\192\006\193\006\194\006\195\006\
    \136\006\196\006\197\006\198\006\199\006\201\006\202\006\203\006\
    \204\006\205\006\206\006\207\006\208\006\209\006\210\006\211\006\
    \189\006\190\006\208\006\212\006\213\006\214\006\210\006\210\006\
    \215\006\216\006\217\006\210\006\218\006\219\006\213\006\220\006\
    \221\006\222\006\223\006\224\006\225\006\226\006\227\006\228\006\
    \229\006\230\006\231\006\232\006\233\006\234\006\235\006\215\006\
    \216\006\236\006\237\006\238\006\239\006\240\006\241\006\242\006\
    \243\006\245\006\246\006\247\006\249\006\215\006\216\006\248\006\
    \190\006\248\006\250\006\251\006\252\006\254\006\255\006\254\006\
    \001\007\003\007\253\006\250\006\250\006\253\006\000\007\243\006\
    \244\006\000\007\253\006\255\006\253\006\248\006\254\006\254\006\
    \002\007\001\007\004\007\002\007\007\007\008\007\001\007\002\007\
    \005\007\006\007\004\007\002\007\009\007\010\007\002\007\244\006\
    \006\007\011\007\005\007\012\007\013\007\014\007\004\007\015\007\
    \016\007\017\007\006\007\018\007\019\007\020\007\021\007\022\007\
    \023\007\006\007\024\007\025\007\028\007\029\007\026\007\030\007\
    \031\007\032\007\033\007\034\007\215\006\216\006\215\006\216\006\
    \244\006\244\006\244\006\244\006\244\006\244\006\244\006\244\006\
    \244\006\027\007\035\007\244\006\244\006\026\007\244\006\244\006\
    \244\006\244\006\244\006\244\006\036\007\244\006\036\007\037\007\
    \038\007\039\007\040\007\041\007\042\007\043\007\037\007\044\007\
    \027\007\045\007\046\007\047\007\048\007\049\007\050\007\051\007\
    \052\007\053\007\054\007\055\007\056\007\057\007\058\007\059\007\
    \060\007\061\007\062\007\063\007\064\007\065\007\066\007\068\007\
    \069\007\067\007\071\007\072\007\070\007\073\007\069\007\074\007\
    \075\007\076\007\077\007\067\007\078\007\079\007\068\007\060\007\
    \061\007\067\007\070\007\080\007\081\007\082\007\027\007\083\007\
    \084\007\085\007\086\007\087\007\088\007\089\007\090\007\075\007\
    \076\007\091\007\092\007\093\007\094\007\095\007\096\007\097\007\
    \098\007\099\007\100\007\101\007\102\007\103\007\104\007\105\007\
    \106\007\107\007\108\007\109\007\106\007\110\007\109\007\111\007\
    \091\007\092\007\112\007\113\007\114\007\115\007\116\007\117\007\
    \118\007\111\007\119\007\061\007\120\007\121\007\122\007\123\007\
    \124\007\125\007\126\007\127\007\076\007\128\007\129\007\215\006\
    \216\006\130\007\109\007\131\007\132\007\133\007\134\007\135\007\
    \136\007\137\007\138\007\109\007\139\007\140\007\092\007\109\007\
    \141\007\142\007\143\007\144\007\145\007\146\007\147\007\148\007\
    \149\007\150\007\151\007\152\007\153\007\154\007\155\007\156\007\
    \157\007\158\007\159\007\160\007\161\007\162\007\163\007\165\007\
    \164\007\166\007\167\007\162\007\164\007\168\007\169\007\170\007\
    \171\007\172\007\173\007\166\007\174\007\175\007\176\007\177\007\
    \178\007\179\007\180\007\181\007\182\007\183\007\184\007\185\007\
    \186\007\187\007\188\007\189\007\190\007\191\007\192\007\193\007\
    \194\007\195\007\196\007\197\007\198\007\199\007\200\007\201\007\
    \202\007\203\007\204\007\205\007\206\007\207\007\208\007\209\007\
    \210\007\208\007\211\007\212\007\213\007\214\007\215\007\216\007\
    \217\007\218\007\219\007\198\007\199\007\220\007\221\007\222\007\
    \223\007\224\007\225\007\226\007\227\007\228\007\229\007\230\007\
    \231\007\232\007\233\007\234\007\235\007\236\007\235\007\237\007\
    \238\007\239\007\240\007\241\007\220\007\221\007\242\007\243\007\
    \244\007\245\007\246\007\247\007\248\007\249\007\250\007\251\007\
    \252\007\253\007\254\007\255\007\000\008\199\007\001\008\002\008\
    \003\008\004\008\005\008\006\008\007\008\008\008\009\008\010\008\
    \011\008\012\008\013\008\014\008\015\008\016\008\017\008\018\008\
    \019\008\020\008\004\008\021\008\022\008\023\008\221\007\024\008\
    \025\008\026\008\221\007\027\008\028\008\029\008\030\008\031\008\
    \032\008\033\008\018\008\034\008\035\008\036\008\037\008\038\008\
    \039\008\040\008\041\008\042\008\043\008\044\008\045\008\046\008\
    \047\008\048\008\049\008\050\008\050\008\051\008\052\008\053\008\
    \043\008\054\008\052\008\055\008\056\008\057\008\058\008\059\008\
    \060\008\061\008\062\008\050\008\063\008\064\008\065\008\066\008\
    \067\008\068\008\069\008\070\008\071\008\072\008\073\008\074\008\
    \075\008\076\008\077\008\078\008\079\008\080\008\081\008\082\008\
    \083\008\084\008\085\008\086\008\087\008\088\008\089\008\090\008\
    \091\008\069\008\070\008\090\008\092\008\093\008\094\008\095\008\
    \096\008\097\008\098\008\099\008\100\008\101\008\102\008\103\008\
    \104\008\105\008\106\008\107\008\108\008\109\008\110\008\111\008\
    \112\008\113\008\114\008\115\008\116\008\117\008\118\008\119\008\
    \120\008\121\008\122\008\123\008\124\008\125\008\126\008\127\008\
    \128\008\129\008\131\008\130\008\132\008\124\008\130\008\133\008\
    \134\008\135\008\136\008\137\008\138\008\139\008\070\008\140\008\
    \141\008\142\008\143\008\144\008\145\008\146\008\147\008\148\008\
    \149\008\150\008\151\008\152\008\153\008\154\008\155\008\156\008\
    \157\008\158\008\159\008\160\008\161\008\162\008\163\008\164\008\
    \165\008\166\008\167\008\168\008\169\008\170\008\171\008\172\008\
    \173\008\178\008\152\008\153\008\165\008\174\008\174\008\179\008\
    \199\008\174\008\180\008\180\008\181\008\181\008\180\008\200\008\
    \181\008\182\008\201\008\183\008\183\008\184\008\184\008\183\008\
    \190\008\184\008\185\008\208\008\174\008\210\008\186\008\186\008\
    \192\008\180\008\186\008\181\008\219\008\220\008\174\008\188\008\
    \182\008\174\008\183\008\191\008\184\008\189\008\180\008\190\008\
    \181\008\185\008\182\008\222\008\183\008\186\008\153\008\183\008\
    \174\008\184\008\195\008\185\008\187\008\187\008\188\008\186\008\
    \192\008\198\008\186\008\204\008\189\008\194\008\194\008\211\008\
    \188\008\194\008\191\008\191\008\196\008\196\008\189\008\205\008\
    \196\008\195\008\205\008\187\008\206\008\203\008\203\008\206\008\
    \198\008\203\008\204\008\174\008\194\008\187\008\211\008\214\008\
    \223\008\221\008\224\008\196\008\227\008\198\008\225\008\225\008\
    \229\008\194\008\226\008\211\008\203\008\238\008\240\008\222\008\
    \196\008\228\008\236\008\228\008\245\008\236\008\214\008\250\008\
    \230\008\203\008\255\008\215\008\244\008\244\008\215\008\196\008\
    \227\008\001\009\003\009\214\008\221\008\007\009\009\009\001\009\
    \015\009\017\009\007\009\018\009\025\009\026\009\027\009\028\009\
    \029\009\255\008\030\009\215\008\215\008\215\008\215\008\031\009\
    \215\008\215\008\223\008\215\008\225\008\224\008\196\008\226\008\
    \205\008\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\034\009\235\008\215\008\041\009\235\008\
    \053\009\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\230\008\215\008\039\009\
    \215\008\043\009\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\215\008\215\008\215\008\
    \215\008\215\008\215\008\215\008\215\008\243\008\243\008\246\008\
    \246\008\243\008\056\009\246\008\247\008\247\008\248\008\248\008\
    \247\008\057\009\248\008\067\009\174\008\235\008\076\009\094\009\
    \095\009\249\008\249\008\099\009\243\008\249\008\246\008\104\009\
    \085\009\107\009\108\009\247\008\243\008\248\008\246\008\109\009\
    \110\009\022\009\111\009\247\008\022\009\248\008\192\008\105\009\
    \249\008\032\009\032\009\035\009\035\009\032\009\112\009\035\009\
    \249\008\191\008\048\009\048\009\013\009\116\009\048\009\013\009\
    \120\009\049\009\049\009\013\009\049\009\049\009\105\009\022\009\
    \032\009\022\009\035\009\187\008\118\009\126\009\130\009\118\009\
    \198\008\048\009\083\009\083\009\063\009\063\009\211\008\063\009\
    \063\009\084\009\089\009\196\008\137\009\022\009\072\009\072\009\
    \135\009\072\009\072\009\141\009\203\008\084\009\145\009\158\009\
    \161\009\083\009\039\009\164\009\043\009\149\009\214\008\166\009\
    \091\009\218\008\091\009\124\009\167\009\001\009\170\009\135\009\
    \007\009\124\009\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\149\009\173\009\176\009\179\009\
    \180\009\215\008\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\218\008\182\009\218\008\
    \185\009\218\008\186\009\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\218\008\218\008\218\008\
    \218\008\218\008\218\008\218\008\218\008\231\008\097\009\097\009\
    \154\009\154\009\097\009\154\009\154\009\194\009\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \190\009\190\009\207\009\190\009\190\009\097\009\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\213\009\231\008\232\009\231\008\085\009\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\013\009\046\009\046\009\236\009\168\009\046\009\022\009\
    \068\009\068\009\068\009\068\009\068\009\068\009\068\009\068\009\
    \068\009\068\009\068\009\068\009\068\009\068\009\068\009\068\009\
    \240\009\046\009\203\009\203\009\168\009\203\009\203\009\084\009\
    \089\009\046\009\243\009\046\009\070\009\070\009\070\009\070\009\
    \070\009\070\009\070\009\070\009\070\009\070\009\070\009\070\009\
    \091\009\083\009\218\008\124\009\209\009\209\009\244\009\209\009\
    \209\009\238\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\046\009\247\009\046\009\248\009\
    \046\009\168\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\046\009\046\009\046\009\046\009\
    \046\009\046\009\046\009\046\009\071\009\071\009\071\009\071\009\
    \071\009\071\009\071\009\071\009\071\009\071\009\249\009\251\009\
    \253\009\001\010\002\010\003\010\006\010\002\010\231\008\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\050\009\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\050\009\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\050\009\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\050\009\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\050\009\050\009\
    \050\009\050\009\050\009\050\009\050\009\050\009\008\010\010\010\
    \012\010\015\010\040\010\046\009\046\009\080\009\080\009\080\009\
    \080\009\080\009\080\009\080\009\080\009\080\009\080\009\046\009\
    \046\009\041\010\051\009\051\009\051\009\051\009\051\009\051\009\
    \051\009\051\009\051\009\051\009\051\009\051\009\051\009\051\009\
    \051\009\051\009\051\009\046\009\051\009\051\009\051\009\051\009\
    \051\009\051\009\051\009\051\009\051\009\051\009\051\009\051\009\
    \051\009\051\009\051\009\051\009\051\009\051\009\051\009\051\009\
    \051\009\051\009\051\009\051\009\051\009\051\009\051\009\051\009\
    \051\009\046\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\238\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\038\010\052\009\052\009\052\009\052\009\052\009\052\009\
    \052\009\052\009\054\009\100\009\100\009\254\009\052\010\100\009\
    \062\010\054\009\071\010\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\230\009\234\009\077\010\
    \255\255\255\255\100\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\013\010\255\255\
    \255\255\013\010\054\009\255\255\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\054\009\054\009\
    \054\009\054\009\054\009\054\009\054\009\054\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \255\255\255\255\255\255\255\255\062\009\255\255\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \062\009\062\009\062\009\062\009\062\009\062\009\062\009\062\009\
    \079\009\079\009\079\009\079\009\079\009\079\009\079\009\079\009\
    \079\009\079\009\079\009\079\009\255\255\054\009\054\009\048\010\
    \048\010\255\255\048\010\048\010\114\009\114\009\255\255\255\255\
    \114\009\054\009\054\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\077\009\077\009\077\009\077\009\077\009\
    \077\009\077\009\077\009\114\009\255\255\054\009\117\009\117\009\
    \255\255\255\255\117\009\128\009\128\009\131\009\131\009\128\009\
    \255\255\131\009\136\009\136\009\255\255\255\255\136\009\139\009\
    \139\009\038\010\255\255\139\009\255\255\117\009\255\255\062\009\
    \062\009\255\255\128\009\254\009\131\009\255\255\142\009\142\009\
    \255\255\136\009\142\009\062\009\062\009\255\255\139\009\201\009\
    \201\009\201\009\201\009\201\009\201\009\201\009\201\009\201\009\
    \201\009\201\009\201\009\230\009\234\009\142\009\255\255\062\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\064\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\064\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\064\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\064\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\064\009\
    \064\009\064\009\064\009\064\009\064\009\064\009\064\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \255\255\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\065\009\065\009\
    \065\009\065\009\065\009\065\009\065\009\065\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\255\255\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\255\255\066\009\066\009\
    \066\009\066\009\066\009\066\009\066\009\066\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\069\009\069\009\
    \069\009\069\009\069\009\069\009\069\009\069\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\073\009\073\009\073\009\
    \073\009\073\009\073\009\073\009\073\009\074\009\074\009\074\009\
    \074\009\074\009\074\009\074\009\074\009\074\009\074\009\074\009\
    \074\009\074\009\074\009\074\009\074\009\074\009\255\255\074\009\
    \074\009\074\009\074\009\074\009\074\009\074\009\074\009\074\009\
    \074\009\074\009\074\009\074\009\074\009\074\009\074\009\074\009\
    \074\009\074\009\074\009\074\009\074\009\074\009\074\009\074\009\
    \074\009\074\009\074\009\074\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\255\255\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\255\255\075\009\075\009\075\009\075\009\
    \075\009\075\009\075\009\075\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\078\009\078\009\078\009\078\009\
    \078\009\078\009\078\009\078\009\081\009\081\009\255\255\147\009\
    \081\009\086\009\086\009\255\255\242\009\086\009\087\009\087\009\
    \101\009\101\009\018\010\090\010\101\009\018\010\133\009\255\255\
    \255\255\245\009\255\255\081\009\019\010\081\009\147\009\019\010\
    \086\009\024\010\086\009\242\009\024\010\087\009\246\009\101\009\
    \044\010\081\009\101\009\101\009\101\009\133\009\086\009\101\009\
    \245\009\101\009\146\009\146\009\255\255\133\009\146\009\255\255\
    \165\009\165\009\250\009\133\009\165\009\246\009\255\255\044\010\
    \255\255\171\009\171\009\174\009\174\009\171\009\255\255\174\009\
    \042\010\146\009\183\009\183\009\187\009\187\009\183\009\165\009\
    \187\009\250\009\255\255\169\009\169\009\150\009\150\009\169\009\
    \171\009\150\009\174\009\250\009\252\009\252\009\255\255\042\010\
    \252\009\183\009\101\009\187\009\255\255\255\255\255\255\255\255\
    \242\009\042\010\169\009\255\255\150\009\255\255\150\009\045\010\
    \054\010\009\010\009\010\252\009\150\009\009\010\255\255\255\255\
    \255\255\255\255\150\009\202\009\202\009\202\009\202\009\202\009\
    \202\009\202\009\202\009\202\009\202\009\246\009\045\010\054\010\
    \009\010\255\255\133\009\255\255\165\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\169\009\
    \255\255\150\009\255\255\150\009\092\010\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\155\009\093\010\255\255\
    \255\255\181\009\181\009\081\009\226\009\181\009\147\009\255\255\
    \086\009\025\010\011\010\011\010\025\010\087\009\011\010\101\009\
    \018\010\090\010\255\255\255\255\255\255\133\009\150\009\150\009\
    \181\009\255\255\019\010\226\009\255\255\226\009\255\255\024\010\
    \255\255\011\010\150\009\150\009\255\255\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\150\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\150\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\255\255\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\255\255\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\159\009\058\010\058\010\
    \255\255\058\010\058\010\255\255\159\009\255\255\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \067\010\067\010\092\010\067\010\067\010\255\255\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\255\255\255\255\255\255\255\255\159\009\255\255\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\159\009\159\009\
    \159\009\255\255\162\009\162\009\073\010\073\010\162\009\073\010\
    \073\010\255\255\255\255\255\255\093\010\255\255\255\255\255\255\
    \181\009\255\255\255\255\226\009\255\255\255\255\255\255\025\010\
    \255\255\162\009\216\009\216\009\216\009\216\009\216\009\216\009\
    \216\009\216\009\216\009\216\009\216\009\216\009\255\255\162\009\
    \199\009\199\009\199\009\199\009\199\009\199\009\199\009\199\009\
    \199\009\199\009\199\009\199\009\199\009\199\009\199\009\199\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \214\009\214\009\214\009\214\009\214\009\214\009\214\009\214\009\
    \159\009\159\009\255\255\255\255\255\255\255\255\027\010\255\255\
    \160\009\027\010\255\255\255\255\159\009\159\009\255\255\160\009\
    \255\255\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\255\255\255\255\255\255\255\255\255\255\
    \159\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\255\255\255\255\255\255\255\255\
    \160\009\255\255\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\255\255\255\255\
    \255\255\255\255\189\009\255\255\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\189\009\189\009\
    \189\009\189\009\189\009\189\009\189\009\189\009\217\009\217\009\
    \217\009\217\009\217\009\217\009\217\009\217\009\217\009\217\009\
    \255\255\255\255\255\255\160\009\160\009\221\009\221\009\221\009\
    \221\009\221\009\221\009\221\009\221\009\221\009\221\009\160\009\
    \160\009\162\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\218\009\218\009\218\009\218\009\218\009\218\009\
    \218\009\218\009\255\255\160\009\220\009\220\009\220\009\220\009\
    \220\009\220\009\220\009\220\009\220\009\220\009\220\009\220\009\
    \224\009\224\009\224\009\224\009\224\009\224\009\224\009\224\009\
    \224\009\224\009\224\009\224\009\255\255\189\009\189\009\225\009\
    \225\009\225\009\225\009\225\009\225\009\225\009\225\009\225\009\
    \225\009\189\009\189\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\222\009\222\009\222\009\222\009\222\009\
    \222\009\222\009\222\009\255\255\027\010\189\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\191\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\191\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\191\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\191\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\191\009\191\009\191\009\
    \191\009\191\009\191\009\191\009\191\009\192\009\192\009\192\009\
    \192\009\192\009\192\009\192\009\192\009\192\009\192\009\192\009\
    \192\009\192\009\192\009\192\009\192\009\192\009\255\255\192\009\
    \192\009\192\009\192\009\192\009\192\009\192\009\192\009\192\009\
    \192\009\192\009\192\009\192\009\192\009\192\009\192\009\192\009\
    \192\009\192\009\192\009\192\009\192\009\192\009\192\009\192\009\
    \192\009\192\009\192\009\192\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\255\255\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\255\255\193\009\193\009\193\009\193\009\
    \193\009\193\009\193\009\193\009\195\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\195\009\255\255\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \255\255\255\255\255\255\255\255\195\009\255\255\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \195\009\195\009\195\009\195\009\195\009\195\009\195\009\195\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \200\009\200\009\200\009\200\009\200\009\200\009\200\009\200\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\195\009\
    \195\009\255\255\255\255\255\255\255\255\255\255\255\255\196\009\
    \255\255\255\255\255\255\195\009\195\009\255\255\196\009\255\255\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\255\255\255\255\255\255\255\255\255\255\195\009\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\030\010\255\255\255\255\030\010\196\009\
    \255\255\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\196\009\196\009\196\009\196\009\196\009\
    \196\009\196\009\196\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\204\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\204\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\204\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\204\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\204\009\204\009\204\009\204\009\204\009\204\009\
    \204\009\204\009\004\010\004\010\028\010\020\010\004\010\028\010\
    \020\010\043\010\065\010\065\010\065\010\065\010\065\010\065\010\
    \065\010\065\010\065\010\065\010\065\010\065\010\255\255\255\255\
    \255\255\004\010\196\009\196\009\255\255\255\255\255\255\255\255\
    \043\010\255\255\197\009\255\255\255\255\255\255\196\009\196\009\
    \020\010\197\009\043\010\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\255\255\255\255\255\255\
    \255\255\255\255\196\009\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\004\010\028\010\
    \255\255\255\255\197\009\255\255\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\197\009\197\009\
    \197\009\197\009\197\009\197\009\197\009\197\009\205\009\205\009\
    \205\009\205\009\205\009\205\009\205\009\205\009\205\009\205\009\
    \205\009\205\009\205\009\205\009\205\009\205\009\205\009\255\255\
    \205\009\205\009\205\009\205\009\205\009\205\009\205\009\205\009\
    \205\009\205\009\205\009\205\009\205\009\205\009\205\009\205\009\
    \205\009\205\009\205\009\205\009\205\009\205\009\205\009\205\009\
    \205\009\205\009\205\009\205\009\205\009\255\255\255\255\032\010\
    \021\010\030\010\032\010\021\010\063\010\063\010\063\010\063\010\
    \063\010\063\010\063\010\063\010\063\010\063\010\063\010\063\010\
    \063\010\063\010\063\010\063\010\255\255\197\009\197\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\198\009\255\255\255\255\
    \255\255\197\009\197\009\021\010\198\009\255\255\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \255\255\255\255\255\255\255\255\255\255\197\009\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\004\010\028\010\020\010\255\255\198\009\255\255\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\198\009\198\009\198\009\198\009\198\009\198\009\198\009\
    \198\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \255\255\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \255\255\206\009\206\009\206\009\206\009\206\009\206\009\206\009\
    \206\009\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \198\009\198\009\033\010\255\255\255\255\033\010\255\255\255\255\
    \208\009\255\255\255\255\255\255\198\009\198\009\255\255\208\009\
    \255\255\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\255\255\255\255\255\255\255\255\255\255\
    \198\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\255\255\032\010\021\010\255\255\
    \208\009\255\255\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\208\009\208\009\208\009\208\009\
    \208\009\208\009\208\009\208\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\210\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\210\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\210\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\210\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\210\009\210\009\210\009\210\009\210\009\
    \210\009\210\009\210\009\066\010\066\010\066\010\066\010\066\010\
    \066\010\066\010\066\010\066\010\066\010\080\010\080\010\080\010\
    \080\010\080\010\080\010\080\010\080\010\080\010\080\010\080\010\
    \080\010\255\255\255\255\208\009\208\009\081\010\081\010\081\010\
    \081\010\081\010\081\010\081\010\081\010\081\010\081\010\208\009\
    \208\009\255\255\211\009\211\009\211\009\211\009\211\009\211\009\
    \211\009\211\009\211\009\211\009\211\009\211\009\211\009\211\009\
    \211\009\211\009\211\009\208\009\211\009\211\009\211\009\211\009\
    \211\009\211\009\211\009\211\009\211\009\211\009\211\009\211\009\
    \211\009\211\009\211\009\211\009\211\009\211\009\211\009\211\009\
    \211\009\211\009\211\009\211\009\211\009\211\009\211\009\211\009\
    \211\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \033\010\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \255\255\212\009\212\009\212\009\212\009\212\009\212\009\212\009\
    \212\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\215\009\215\009\215\009\215\009\215\009\215\009\215\009\
    \215\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\219\009\219\009\219\009\219\009\219\009\219\009\219\009\
    \219\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\223\009\223\009\223\009\223\009\223\009\223\009\223\009\
    \223\009\016\010\016\010\022\010\055\010\016\010\022\010\026\010\
    \026\010\031\010\031\010\026\010\035\010\031\010\036\010\035\010\
    \037\010\036\010\255\255\037\010\255\255\255\255\255\255\255\255\
    \016\010\255\255\255\255\055\010\255\255\255\255\026\010\255\255\
    \031\010\255\255\255\255\255\255\255\255\016\010\022\010\084\010\
    \084\010\084\010\084\010\084\010\084\010\084\010\084\010\084\010\
    \084\010\084\010\084\010\255\255\255\255\055\010\085\010\085\010\
    \085\010\085\010\085\010\085\010\085\010\085\010\085\010\085\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\255\255\026\010\255\255\031\010\046\010\035\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\046\010\046\010\046\010\046\010\046\010\046\010\
    \046\010\046\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\049\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\049\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\049\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\049\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\049\010\049\010\049\010\049\010\049\010\049\010\049\010\
    \049\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\078\010\078\010\078\010\078\010\078\010\078\010\078\010\
    \078\010\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\046\010\046\010\089\010\089\010\089\010\089\010\089\010\
    \089\010\089\010\089\010\089\010\089\010\046\010\046\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\082\010\
    \082\010\082\010\082\010\082\010\082\010\082\010\082\010\255\255\
    \255\255\046\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\086\010\086\010\086\010\086\010\086\010\086\010\
    \086\010\086\010\255\255\255\255\255\255\255\255\255\255\255\255\
    \016\010\022\010\255\255\255\255\255\255\255\255\026\010\046\010\
    \031\010\255\255\035\010\255\255\036\010\255\255\037\010\050\010\
    \050\010\050\010\050\010\050\010\050\010\050\010\050\010\050\010\
    \050\010\050\010\050\010\050\010\050\010\050\010\050\010\050\010\
    \255\255\050\010\050\010\050\010\050\010\050\010\050\010\050\010\
    \050\010\050\010\050\010\050\010\050\010\050\010\050\010\050\010\
    \050\010\050\010\050\010\050\010\050\010\050\010\050\010\050\010\
    \050\010\050\010\050\010\050\010\050\010\050\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\255\255\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\053\010\051\010\051\010\
    \051\010\051\010\051\010\051\010\051\010\051\010\088\010\088\010\
    \088\010\088\010\088\010\088\010\088\010\088\010\088\010\088\010\
    \088\010\088\010\255\255\255\255\053\010\255\255\255\255\255\255\
    \255\255\255\255\255\255\053\010\255\255\255\255\255\255\255\255\
    \255\255\255\255\053\010\255\255\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\255\255\
    \255\255\255\255\255\255\255\255\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\255\255\
    \255\255\255\255\255\255\053\010\255\255\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\053\010\
    \053\010\053\010\053\010\053\010\053\010\053\010\053\010\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\255\255\255\255\255\255\255\255\056\010\255\255\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\056\010\056\010\056\010\056\010\056\010\056\010\056\010\
    \056\010\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\053\010\053\010\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\053\010\053\010\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\053\010\255\255\
    \255\255\255\255\255\255\057\010\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \056\010\056\010\057\010\255\255\255\255\255\255\255\255\255\255\
    \255\255\057\010\255\255\255\255\056\010\056\010\255\255\255\255\
    \057\010\255\255\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\255\255\255\255\255\255\
    \056\010\255\255\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\255\255\255\255\255\255\
    \255\255\057\010\255\255\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\057\010\057\010\057\010\
    \057\010\057\010\057\010\057\010\057\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\059\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\059\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\059\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\059\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\059\010\059\010\059\010\059\010\
    \059\010\059\010\059\010\059\010\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\057\010\057\010\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \057\010\057\010\255\255\060\010\060\010\060\010\060\010\060\010\
    \060\010\060\010\060\010\060\010\060\010\060\010\060\010\060\010\
    \060\010\060\010\060\010\060\010\057\010\060\010\060\010\060\010\
    \060\010\060\010\060\010\060\010\060\010\060\010\060\010\060\010\
    \060\010\060\010\060\010\060\010\060\010\060\010\060\010\060\010\
    \060\010\060\010\060\010\060\010\060\010\060\010\060\010\060\010\
    \060\010\060\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\255\255\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\255\255\061\010\061\010\061\010\061\010\061\010\061\010\
    \061\010\061\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\064\010\064\010\064\010\064\010\064\010\064\010\
    \064\010\064\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\068\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\068\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\068\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\068\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\068\010\068\010\068\010\068\010\068\010\068\010\068\010\
    \068\010\069\010\069\010\069\010\069\010\069\010\069\010\069\010\
    \069\010\069\010\069\010\069\010\069\010\069\010\069\010\069\010\
    \069\010\069\010\255\255\069\010\069\010\069\010\069\010\069\010\
    \069\010\069\010\069\010\069\010\069\010\069\010\069\010\069\010\
    \069\010\069\010\069\010\069\010\069\010\069\010\069\010\069\010\
    \069\010\069\010\069\010\069\010\069\010\069\010\069\010\069\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\255\255\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\072\010\
    \070\010\070\010\070\010\070\010\070\010\070\010\070\010\070\010\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\072\010\255\255\
    \255\255\255\255\255\255\255\255\255\255\072\010\255\255\255\255\
    \255\255\255\255\255\255\255\255\072\010\255\255\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\255\255\255\255\255\255\255\255\255\255\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\255\255\255\255\255\255\255\255\072\010\255\255\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\072\010\072\010\072\010\072\010\072\010\072\010\072\010\
    \072\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \074\010\074\010\074\010\074\010\074\010\074\010\074\010\074\010\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \072\010\072\010\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\072\010\072\010\255\255\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \072\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\075\010\075\010\
    \075\010\075\010\075\010\075\010\075\010\075\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\255\255\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\255\255\076\010\076\010\
    \076\010\076\010\076\010\076\010\076\010\076\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\079\010\079\010\
    \079\010\079\010\079\010\079\010\079\010\079\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\083\010\083\010\
    \083\010\083\010\083\010\083\010\083\010\083\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\087\010\087\010\
    \087\010\087\010\087\010\087\010\087\010\087\010\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code =
   "";
  Lexing.lex_backtrk_code =
   "";
  Lexing.lex_default_code =
   "";
  Lexing.lex_trans_code =
   "";
  Lexing.lex_check_code =
   "";
  Lexing.lex_code =
   "";
}

let rec coq_bol lexbuf =
   __ocaml_lex_coq_bol_rec lexbuf 0
and __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 443 "tools/coqdoc/cpretty.mll"
      ( if not (!in_proof <> None && (!Cdglobals.gallina || !Cdglobals.light))
        then Output.empty_line_of_code ();
        coq_bol lexbuf )
# 3329 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 447 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3337 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 452 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc (); comments lexbuf; Output.end_doc ();
	Output.start_coq (); coq lexbuf )
# 3343 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 455 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf; coq_bol lexbuf )
# 3348 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 457 "tools/coqdoc/cpretty.mll"
      ( begin_show (); coq_bol lexbuf )
# 3353 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 459 "tools/coqdoc/cpretty.mll"
      ( end_show (); coq_bol lexbuf )
# 3358 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 461 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    let eol = skip_to_dot lexbuf in
	      if eol then (coq_bol lexbuf) else coq lexbuf
	  else
	    begin
	      output_indented_keyword s lexbuf;
	      let eol = body lexbuf in
	      if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3372 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 472 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
        output_indented_keyword s lexbuf;
        let eol = body lexbuf in
	in_proof := Some eol;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3381 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 478 "tools/coqdoc/cpretty.mll"
      ( in_proof := Some true;
	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    let s = lexeme lexbuf in
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3394 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 487 "tools/coqdoc/cpretty.mll"
                  (
      let eol =
	if not (only_gallina ()) then
	  begin backtrack lexbuf; body_bol lexbuf end
	else skip_to_dot lexbuf
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3406 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 496 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3416 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 503 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3426 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 510 "tools/coqdoc/cpretty.mll"
      (	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3434 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 516 "tools/coqdoc/cpretty.mll"
      ( let tok = lexeme lexbuf in
	let s = printing_token_body lexbuf in
	  add_printing_token tok s;
	  coq_bol lexbuf )
# 3442 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 521 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3451 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 528 "tools/coqdoc/cpretty.mll"
      ( remove_printing_token (lexeme lexbuf);
	coq_bol lexbuf )
# 3457 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 531 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'remove printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3466 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 537 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        let eol =
          if parse_comments () then begin
            let s = lexeme lexbuf in
            let nbsp, isp = count_spaces s in
            Output.indentation nbsp;
            Output.start_comment ();
            comment lexbuf
          end else skipped_comment lexbuf in
        if eol then coq_bol lexbuf else coq lexbuf )
# 3480 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 548 "tools/coqdoc/cpretty.mll"
      ( () )
# 3485 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 550 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    skip_to_dot_or_brace lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3496 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state

and coq lexbuf =
   __ocaml_lex_coq_rec lexbuf 1159
and __ocaml_lex_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 562 "tools/coqdoc/cpretty.mll"
      ( if not (only_gallina ()) then Output.line_break(); coq_bol lexbuf )
# 3508 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 564 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3516 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 569 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        let eol =
          if parse_comments () then begin
            Output.start_comment ();
            comment lexbuf
          end else skipped_comment lexbuf in
        if eol then coq_bol lexbuf else coq lexbuf )
# 3527 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 577 "tools/coqdoc/cpretty.mll"
      ( if not !formatted then
	begin
	  (* Isn't this an anomaly *)
	  let s = lexeme lexbuf in
	  let nlsp,s = remove_newline s in
	  let nbsp,isp = count_spaces s in
	  Output.indentation nbsp;
	  let loc = lexeme_start lexbuf + isp + nlsp in
	  Output.sublexer ']' loc;
	  Output.sublexer ']' (loc+1);
	  coq lexbuf
	end )
# 3543 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 590 "tools/coqdoc/cpretty.mll"
      ( () )
# 3548 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 592 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    begin
	      let eol = skip_to_dot lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end
	  else
	    begin
	      Output.ident s None;
	      let eol=body lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3564 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 605 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    let s = lexeme lexbuf in
	    let eol =
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	    in
	      eol
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3579 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 616 "tools/coqdoc/cpretty.mll"
           (
      let eol =
	if not !Cdglobals.gallina then
	  begin backtrack lexbuf; body lexbuf end
	else
	  let eol = skip_to_dot lexbuf in
	    if !in_proof <> Some true && eol then
	      Output.line_break ();
	    eol
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3595 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 629 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3603 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 634 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	Output.ident s None;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3611 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 639 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3619 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 643 "tools/coqdoc/cpretty.mll"
           ( Output.char ' '; coq lexbuf )
# 3624 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 645 "tools/coqdoc/cpretty.mll"
      ( () )
# 3629 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 646 "tools/coqdoc/cpretty.mll"
      (	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    skip_to_dot_or_brace lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf)
# 3640 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_coq_rec lexbuf __ocaml_lex_state

and doc_bol lexbuf =
   __ocaml_lex_doc_bol_rec lexbuf 2222
and __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 658 "tools/coqdoc/cpretty.mll"
      ( let eol, lex = strip_eol (lexeme lexbuf) in
        let lev, s = sec_title lex in
          if (!Cdglobals.lib_subtitles) &&
             (subtitle (Output.get_module false) s) then
            ()
          else
            Output.section lev (fun () -> ignore (doc None (from_string s)));
	    if eol then doc_bol lexbuf else doc None lexbuf )
# 3659 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 667 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
        let lines = (List.length bufs) - 1 in
        let line = 
          match bufs with
          | [] -> eprintf "Internal error bad_split1 - please report\n";
                  exit 1
          | _ -> List.nth bufs lines
        in
          match check_start_list line with
          | Neither -> backtrack_past_newline lexbuf; doc None lexbuf
          | List n -> if lines > 0 then Output.paragraph ();
                      Output.item 1; doc (Some [n]) lexbuf
          | Rule -> Output.rule (); doc None lexbuf
      )
# 3678 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 683 "tools/coqdoc/cpretty.mll"
      ( Output.paragraph (); doc_bol lexbuf )
# 3683 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 685 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false; verbatim 0 false lexbuf; doc_bol lexbuf )
# 3688 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 687 "tools/coqdoc/cpretty.mll"
      ( true )
# 3693 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 689 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else start_emph ();
        doc None lexbuf )
# 3699 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 692 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; doc None lexbuf )
# 3704 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state

and doc_list_bol indents lexbuf =
   __ocaml_lex_doc_list_bol_rec indents lexbuf 2244
and __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 697 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> backtrack lexbuf; doc_bol lexbuf
        | StartLevel n -> Output.item n; doc (Some (take n indents)) lexbuf
        | InLevel (n,true) ->
            let items = List.length indents in
              Output.item (items+1);
              doc (Some (List.append indents [n_spaces])) lexbuf
        | InLevel (_,false) ->
            backtrack lexbuf; doc_bol lexbuf
      )
# 3726 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 709 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false;
        verbatim 0 false lexbuf;
        doc_list_bol indents lexbuf )
# 3733 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 713 "tools/coqdoc/cpretty.mll"
      ( formatted := true;
        Output.start_inline_coq_block ();
        ignore(body_bol lexbuf);
        Output.end_inline_coq_block ();
        formatted := false;
        doc_list_bol indents lexbuf )
# 3743 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 720 "tools/coqdoc/cpretty.mll"
      ( inf_rules (Some indents) lexbuf )
# 3748 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 722 "tools/coqdoc/cpretty.mll"
      ( (* Like in the doc_bol production, these two productions
           exist only to deal properly with whitespace *)
        Output.paragraph ();
        backtrack_past_newline lexbuf;
        doc_list_bol indents lexbuf )
# 3757 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 728 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let buf =
          let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
            match bufs with
              | (_ :: s :: []) -> s
              | (_ :: _ :: s :: _) -> s
              | _ -> eprintf "Internal error bad_split2 - please report\n";
                     exit 1
        in
        let (n_spaces,_) = count_spaces buf in
        match find_level indents n_spaces with
        | StartLevel 1 | Before ->
        (* Here we were at the beginning of a line, and it was blank.
           The next line started before any list items.  So: insert
           a paragraph for the empty line, rewind to whatever's just
           after the newline, then toss over to doc_bol for whatever
           comes next. *)
            Output.stop_item ();
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_bol lexbuf
        | StartLevel _ | InLevel _ ->
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_list_bol indents lexbuf

      )
# 3788 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 756 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> Output.stop_item (); backtrack lexbuf;
                    doc_bol lexbuf
        | StartLevel n ->
            Output.reach_item_level (n-1);
            backtrack lexbuf;
            doc (Some (take (n-1) indents)) lexbuf
        | InLevel (n,_) ->
            Output.reach_item_level n;
            backtrack lexbuf;
            doc (Some (take n indents)) lexbuf
      )
# 3805 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state

and doc indents lexbuf =
   __ocaml_lex_doc_rec indents lexbuf 2263
and __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 773 "tools/coqdoc/cpretty.mll"
      ( Output.char '\n';
        match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 3820 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 778 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments
        then (Output.char '['; Output.char '['; doc indents lexbuf)
        else (formatted := true;
              Output.start_inline_coq_block ();
	      let eol = body_bol lexbuf in
	        Output.end_inline_coq_block (); formatted := false;
	        if eol then
		  match indents with
		  | Some ls -> doc_list_bol ls lexbuf
		  | None -> doc_bol lexbuf
		else doc indents lexbuf))
# 3835 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 790 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 3840 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 792 "tools/coqdoc/cpretty.mll"
      ( Output.proofbox (); doc indents lexbuf )
# 3845 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 793 "tools/coqdoc/cpretty.mll"
         ( url lexbuf; doc indents lexbuf )
# 3850 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 795 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1;  Output.start_inline_coq (); escaped_coq lexbuf;
              Output.end_inline_coq ()); doc indents lexbuf )
# 3857 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 799 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf ;
        let bol_parse = match indents with
                        | Some is -> doc_list_bol is
                        | None   -> doc_bol
        in
        let eol =
          if !Cdglobals.parse_comments then comment lexbuf
          else skipped_comment lexbuf in
        if eol then bol_parse lexbuf else doc indents lexbuf )
# 3870 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 809 "tools/coqdoc/cpretty.mll"
      ((match indents with
        | Some _ -> Output.stop_item ()
        | None -> ());
       (* this says - if there is a blank line between the two comments,
          insert one in the output too *)
       let lines = List.length (Str.split_delim (Str.regexp "['\n']") 
                                                (lexeme lexbuf))
       in 
         if lines > 2 then Output.paragraph ();
       doc_bol lexbuf
      )
# 3885 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 821 "tools/coqdoc/cpretty.mll"
      ( true )
# 3890 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 823 "tools/coqdoc/cpretty.mll"
      ( false )
# 3895 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 825 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$'
        else (Output.start_latex_math (); escaped_math_latex lexbuf);
        doc indents lexbuf )
# 3902 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 829 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$';
        Output.char '$'; doc indents lexbuf )
# 3908 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 832 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%'
        else escaped_latex lexbuf; doc indents lexbuf )
# 3914 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 835 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%';
        Output.char '%'; doc indents lexbuf )
# 3920 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 838 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#'
        else escaped_html lexbuf; doc indents lexbuf )
# 3926 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 841 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#';
        Output.char '#'; doc indents lexbuf )
# 3932 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 844 "tools/coqdoc/cpretty.mll"
      ( List.iter (fun x -> Output.char (lexeme_char lexbuf x)) [0;1;2];
        doc indents lexbuf)
# 3938 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 847 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0);
        if !Cdglobals.plain_comments then Output.char '_' else  start_emph () ;
        doc indents lexbuf )
# 3945 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 851 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else stop_emph () ;
        Output.char (lexeme_char lexbuf 1);
        doc indents lexbuf )
# 3952 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 855 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim true; verbatim 0 true lexbuf; doc_bol lexbuf )
# 3957 "tools/coqdoc/cpretty.ml"

  | 20 ->
# 857 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments 
	then Output.char '"' 
	else if in_quote ()
	then stop_quote () 
	else start_quote ();
        doc indents lexbuf )
# 3967 "tools/coqdoc/cpretty.ml"

  | 21 ->
# 864 "tools/coqdoc/cpretty.mll"
      ( false )
# 3972 "tools/coqdoc/cpretty.ml"

  | 22 ->
# 866 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); doc indents lexbuf )
# 3977 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state

and escaped_math_latex lexbuf =
   __ocaml_lex_escaped_math_latex_rec lexbuf 2305
and __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 871 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 3989 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 872 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 3994 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 874 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math (); backtrack lexbuf )
# 3999 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 875 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_math_latex lexbuf )
# 4004 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state

and escaped_latex lexbuf =
   __ocaml_lex_escaped_latex_rec lexbuf 2311
and __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 878 "tools/coqdoc/cpretty.mll"
        ( () )
# 4016 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 879 "tools/coqdoc/cpretty.mll"
        ( () )
# 4021 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 881 "tools/coqdoc/cpretty.mll"
        ( backtrack lexbuf )
# 4026 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 882 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_latex lexbuf )
# 4031 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state

and escaped_html lexbuf =
   __ocaml_lex_escaped_html_rec lexbuf 2317
and __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 885 "tools/coqdoc/cpretty.mll"
        ( () )
# 4043 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 887 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '&'; Output.html_char '#'; escaped_html lexbuf )
# 4048 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 889 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '#'; escaped_html lexbuf )
# 4053 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 890 "tools/coqdoc/cpretty.mll"
        ( () )
# 4058 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 892 "tools/coqdoc/cpretty.mll"
        ( backtrack lexbuf )
# 4063 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 893 "tools/coqdoc/cpretty.mll"
        ( Output.html_char (lexeme_char lexbuf 0); escaped_html lexbuf )
# 4068 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state

and verbatim depth inline lexbuf =
   __ocaml_lex_verbatim_rec depth inline lexbuf 2326
and __ocaml_lex_verbatim_rec depth inline lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 896 "tools/coqdoc/cpretty.mll"
                      ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 4080 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 897 "tools/coqdoc/cpretty.mll"
            ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 4085 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 898 "tools/coqdoc/cpretty.mll"
         ( Output.stop_verbatim inline )
# 4090 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 899 "tools/coqdoc/cpretty.mll"
         ( Output.verbatim_char inline '(';
           Output.verbatim_char inline '*';
           verbatim (depth+1) inline lexbuf )
# 4097 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 902 "tools/coqdoc/cpretty.mll"
         ( if (depth == 0)
           then (Output.stop_verbatim inline; backtrack lexbuf)
           else (Output.verbatim_char inline '*';
                 Output.verbatim_char inline ')';
                 verbatim (depth-1) inline lexbuf) )
# 4106 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 907 "tools/coqdoc/cpretty.mll"
        ( Output.stop_verbatim inline )
# 4111 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 908 "tools/coqdoc/cpretty.mll"
      ( Output.verbatim_char inline (lexeme_char lexbuf 0); verbatim depth inline lexbuf )
# 4116 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_verbatim_rec depth inline lexbuf __ocaml_lex_state

and url lexbuf =
   __ocaml_lex_url_rec lexbuf 2343
and __ocaml_lex_url_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 911 "tools/coqdoc/cpretty.mll"
         ( Output.url (Buffer.contents url_buffer) None; Buffer.clear url_buffer )
# 4128 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 912 "tools/coqdoc/cpretty.mll"
        ( url_name lexbuf )
# 4133 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 913 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_buffer (lexeme_char lexbuf 0); url lexbuf )
# 4138 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_url_rec lexbuf __ocaml_lex_state

and url_name lexbuf =
   __ocaml_lex_url_name_rec lexbuf 2347
and __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 916 "tools/coqdoc/cpretty.mll"
        ( Output.url (Buffer.contents url_buffer) (Some (Buffer.contents url_name_buffer));
	  Buffer.clear url_buffer; Buffer.clear url_name_buffer )
# 4151 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 918 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_name_buffer (lexeme_char lexbuf 0); url_name lexbuf )
# 4156 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state

and escaped_coq lexbuf =
   __ocaml_lex_escaped_coq_rec lexbuf 2350
and __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 924 "tools/coqdoc/cpretty.mll"
      ( decr brackets;
	if !brackets > 0 then
	  (Output.sublexer_in_doc ']'; escaped_coq lexbuf)
	else Tokens.flush_sublexer () )
# 4171 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 929 "tools/coqdoc/cpretty.mll"
      ( incr brackets;
        Output.sublexer_in_doc '['; escaped_coq lexbuf )
# 4177 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 932 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer (); comment_level := 1;
        ignore (if !Cdglobals.parse_comments then comment lexbuf
                else skipped_comment lexbuf);
        escaped_coq lexbuf )
# 4185 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 937 "tools/coqdoc/cpretty.mll"
      ( (* likely to be a syntax error: we escape *) backtrack lexbuf )
# 4190 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 939 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer () )
# 4195 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 941 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
        escaped_coq lexbuf )
# 4202 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 945 "tools/coqdoc/cpretty.mll"
      ( let str = lexeme lexbuf in
          Tokens.flush_sublexer();
          (if !Cdglobals.inline_notmono then () 
                                        else Output.end_inline_coq ()); 
          String.iter Output.char str; 
          (if !Cdglobals.inline_notmono then () 
                                        else Output.start_inline_coq ());
          escaped_coq lexbuf )
# 4214 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 954 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer_in_doc (lexeme_char lexbuf 0);
        escaped_coq lexbuf )
# 4220 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state

and comments lexbuf =
   __ocaml_lex_comments_rec lexbuf 2385
and __ocaml_lex_comments_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 961 "tools/coqdoc/cpretty.mll"
      ( Output.char ' '; comments lexbuf )
# 4232 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 963 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	let s = String.sub s 1 (String.length s - 2) in
	ignore (doc None (from_string s)); comments lexbuf )
# 4239 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 967 "tools/coqdoc/cpretty.mll"
      ( escaped_coq (from_string (lexeme lexbuf)); comments lexbuf )
# 4244 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 969 "tools/coqdoc/cpretty.mll"
      ( () )
# 4249 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 971 "tools/coqdoc/cpretty.mll"
      ( () )
# 4254 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 973 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); comments lexbuf )
# 4259 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comments_rec lexbuf __ocaml_lex_state

and skipped_comment lexbuf =
   __ocaml_lex_skipped_comment_rec lexbuf 2395
and __ocaml_lex_skipped_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 977 "tools/coqdoc/cpretty.mll"
      ( incr comment_level;
	skipped_comment lexbuf )
# 4272 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 980 "tools/coqdoc/cpretty.mll"
      ( decr comment_level;
        if !comment_level > 0 then skipped_comment lexbuf else true )
# 4278 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 983 "tools/coqdoc/cpretty.mll"
      ( decr comment_level;
        if !comment_level > 0 then skipped_comment lexbuf else false )
# 4284 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 985 "tools/coqdoc/cpretty.mll"
        ( false )
# 4289 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 986 "tools/coqdoc/cpretty.mll"
      ( skipped_comment lexbuf )
# 4294 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skipped_comment_rec lexbuf __ocaml_lex_state

and comment lexbuf =
   __ocaml_lex_comment_rec lexbuf 2405
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 990 "tools/coqdoc/cpretty.mll"
      ( incr comment_level;
        Output.start_comment ();
        comment lexbuf )
# 4308 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 994 "tools/coqdoc/cpretty.mll"
      ( Output.end_comment ();
        Output.line_break ();
        decr comment_level;
        if !comment_level > 0 then comment lexbuf else true )
# 4316 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 999 "tools/coqdoc/cpretty.mll"
      ( Output.end_comment ();
        decr comment_level;
        if !comment_level > 0 then comment lexbuf else false )
# 4323 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1003 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1; Output.start_inline_coq ();
              escaped_coq lexbuf; Output.end_inline_coq ());
        comment lexbuf )
# 4331 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1008 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then (Output.char '['; Output.char '[')
        else (formatted := true;
              Output.start_inline_coq_block ();
              let _ = body_bol lexbuf in
                Output.end_inline_coq_block (); formatted := false);
        comment lexbuf )
# 4341 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1015 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$'
        else (Output.start_latex_math (); escaped_math_latex lexbuf);
        comment lexbuf )
# 4348 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1019 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$';
        Output.char '$';
        comment lexbuf )
# 4355 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1023 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%'
        else escaped_latex lexbuf;
        comment lexbuf )
# 4362 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1027 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%';
        Output.char '%';
        comment lexbuf )
# 4369 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1031 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#'
        else escaped_html lexbuf;
        comment lexbuf )
# 4376 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1035 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#';
        Output.char '#';
        comment lexbuf )
# 4383 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1038 "tools/coqdoc/cpretty.mll"
         ( false )
# 4388 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1040 "tools/coqdoc/cpretty.mll"
      ( Output.indentation (fst (count_spaces (lexeme lexbuf)));
        comment lexbuf )
# 4394 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1043 "tools/coqdoc/cpretty.mll"
      ( Output.line_break ();
        comment lexbuf )
# 4400 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1045 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0);
        comment lexbuf )
# 4406 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and skip_to_dot lexbuf =
   __ocaml_lex_skip_to_dot_rec lexbuf 2428
and __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1049 "tools/coqdoc/cpretty.mll"
                  ( true )
# 4418 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1050 "tools/coqdoc/cpretty.mll"
                     ( false )
# 4423 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1052 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        ignore (skipped_comment lexbuf);
        skip_to_dot lexbuf )
# 4430 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1055 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot lexbuf )
# 4435 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state

and skip_to_dot_or_brace lexbuf =
   __ocaml_lex_skip_to_dot_or_brace_rec lexbuf 2437
and __ocaml_lex_skip_to_dot_or_brace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1058 "tools/coqdoc/cpretty.mll"
                  ( true )
# 4447 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1059 "tools/coqdoc/cpretty.mll"
                     ( false )
# 4452 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1061 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
        ignore (skipped_comment lexbuf);
        skip_to_dot_or_brace lexbuf )
# 4459 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1065 "tools/coqdoc/cpretty.mll"
      ( true )
# 4464 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1067 "tools/coqdoc/cpretty.mll"
      ( false )
# 4469 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1069 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot_or_brace lexbuf )
# 4474 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1070 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot lexbuf )
# 4479 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_to_dot_or_brace_rec lexbuf __ocaml_lex_state

and body_bol lexbuf =
   __ocaml_lex_body_bol_rec lexbuf 2451
and __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1074 "tools/coqdoc/cpretty.mll"
      ( Output.indentation (fst (count_spaces (lexeme lexbuf))); body lexbuf )
# 4491 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1075 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; Output.indentation 0; body lexbuf )
# 4496 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state

and body lexbuf =
   __ocaml_lex_body_rec lexbuf 2454
and __ocaml_lex_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1078 "tools/coqdoc/cpretty.mll"
       (Tokens.flush_sublexer(); Output.line_break(); Lexing.new_line lexbuf; body_bol lexbuf)
# 4508 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1080 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
            let s = lexeme lexbuf in
            let nlsp,s = remove_newline s in
            let _,isp = count_spaces s in
            let loc = lexeme_start lexbuf + nlsp + isp in
            Output.sublexer ']' loc;
            Output.sublexer ']' (loc+1);
            Tokens.flush_sublexer();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4529 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1098 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
	    let loc = lexeme_start lexbuf in
	    Output.sublexer ']' loc;
	    Output.sublexer ']' (loc+1);
	    Tokens.flush_sublexer();
	    Output.line_break();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4548 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1113 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); false )
# 4553 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1115 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.'; Output.line_break();
	  if not !formatted then true else body_bol lexbuf )
# 4559 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1118 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.';
        if not !formatted then
          begin
            eprintf "Error: stray ]] at %d\n"  (lexeme_start lexbuf);
            flush stderr;
            exit 1
          end
	  else
          begin
            Output.paragraph ();
            true
          end
      )
# 4576 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1132 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); Output.char '.'; Output.char ' ';
	  if not !formatted then false else body lexbuf )
# 4582 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1135 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then body_bol lexbuf else body lexbuf )
# 4590 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1140 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); comment_level := 1;
        let eol =
          if parse_comments () then begin
            Output.start_comment ();
            comment lexbuf
          end else begin
            let eol = skipped_comment lexbuf in
            if eol then Output.line_break();
            eol
          end in
	if eol then body_bol lexbuf else body lexbuf )
# 4605 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1152 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
	start_notation_string lexbuf )
# 4612 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1156 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) (Some (lexeme_start lexbuf));
	body lexbuf )
# 4619 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1160 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '.'; Output.char '.';
        body lexbuf )
# 4625 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1163 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"';
        string lexbuf;
        body lexbuf )
# 4632 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1167 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
        body lexbuf )
# 4638 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1170 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
	Output.sublexer c (lexeme_start lexbuf);
        body lexbuf )
# 4645 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_body_rec lexbuf __ocaml_lex_state

and start_notation_string lexbuf =
   __ocaml_lex_start_notation_string_rec lexbuf 2530
and __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1175 "tools/coqdoc/cpretty.mll"
          ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
	    start_notation_string lexbuf )
# 4658 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1178 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer '"' (lexeme_start lexbuf);
        notation_string lexbuf;
	body lexbuf )
# 4665 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1182 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; body lexbuf )
# 4670 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state

and notation_string lexbuf =
   __ocaml_lex_notation_string_rec lexbuf 2534
and __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1186 "tools/coqdoc/cpretty.mll"
      ( Output.char '"'; Output.char '"'; (* Unlikely! *)
        notation_string lexbuf )
# 4683 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1189 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"' )
# 4688 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1190 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
        Output.sublexer c (lexeme_start lexbuf);
        notation_string lexbuf )
# 4695 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state

and string lexbuf =
   __ocaml_lex_string_rec lexbuf 2538
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1195 "tools/coqdoc/cpretty.mll"
           ( Output.char '"'; Output.char '"'; string lexbuf )
# 4707 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1196 "tools/coqdoc/cpretty.mll"
           ( Output.char '"' )
# 4712 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1197 "tools/coqdoc/cpretty.mll"
           ( let c = lexeme_char lexbuf 0 in Output.char c; string lexbuf )
# 4717 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and skip_hide lexbuf =
   __ocaml_lex_skip_hide_rec lexbuf 2542
and __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1200 "tools/coqdoc/cpretty.mll"
                   ( () )
# 4729 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1201 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf )
# 4734 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state

and printing_token_body lexbuf =
   __ocaml_lex_printing_token_body_rec lexbuf 2558
and __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1207 "tools/coqdoc/cpretty.mll"
 ( let s = Buffer.contents token_buffer in
	  Buffer.clear token_buffer;
	  s )
# 4748 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1210 "tools/coqdoc/cpretty.mll"
        ( Buffer.add_string token_buffer (lexeme lexbuf);
	  printing_token_body lexbuf )
# 4754 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state

and inf_rules indents lexbuf =
   __ocaml_lex_inf_rules_rec indents lexbuf 2564
and __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1217 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 4766 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1219 "tools/coqdoc/cpretty.mll"
      ( match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 4773 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1223 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;  (* anything else must be the first line in a rule *)
        inf_rules_assumptions indents [] lexbuf)
# 4779 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state

and inf_rules_assumptions indents assumptions lexbuf =
   __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf 2576
and __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1232 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let dashes_and_name = 
               cut_head_tail_spaces (String.sub line 0 (String.length line - 1))
        in
        let ldn = String.length dashes_and_name in
        let (dashes,name) = 
          try (let i = String.index dashes_and_name ' ' in
               let d = String.sub dashes_and_name 0 i in
               let n = cut_head_tail_spaces 
                        (String.sub dashes_and_name (i+1) (ldn-i-1))
               in
                 (d, Some n))
          with _ -> (dashes_and_name, None)
            
        in
          inf_rules_conclusion indents (List.rev assumptions) 
                               (spaces, dashes, name) [] lexbuf )
# 4808 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1251 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let assumption = cut_head_tail_spaces 
                            (String.sub line 0 (String.length line - 1))
        in
          inf_rules_assumptions indents ((spaces,assumption)::assumptions)
                                lexbuf )
# 4819 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state

and inf_rules_conclusion indents assumptions middle conclusions lexbuf =
   __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf 2586
and __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1265 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;
        Output.inf_rule assumptions middle (List.rev conclusions);
        inf_rules indents lexbuf )
# 4833 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1269 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let conc = cut_head_tail_spaces (String.sub line 0 
                                                    (String.length line - 1))
        in
          inf_rules_conclusion indents assumptions middle 
                               ((spaces,conc) :: conclusions) lexbuf
      )
# 4845 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state

and st_start m lexbuf =
   __ocaml_lex_st_start_rec m lexbuf 2598
and __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1281 "tools/coqdoc/cpretty.mll"
      ( st_modname m lexbuf )
# 4857 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1283 "tools/coqdoc/cpretty.mll"
      ( None )
# 4862 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state

and st_modname m lexbuf =
   __ocaml_lex_st_modname_rec m lexbuf 2606
and __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1287 "tools/coqdoc/cpretty.mll"
      ( if subtitle m (lexeme lexbuf) then
          st_subtitle lexbuf
        else
          None
      )
# 4878 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1293 "tools/coqdoc/cpretty.mll"
      ( None )
# 4883 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state

and st_subtitle lexbuf =
   __ocaml_lex_st_subtitle_rec lexbuf 2650
and __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1297 "tools/coqdoc/cpretty.mll"
      ( let st = lexeme lexbuf in
        let i = try Str.search_forward (Str.regexp "\\**)") st 0 with
                   Not_found ->
                     (eprintf "unterminated comment at beginning of file\n";
                      exit 1)
        in
          Some (cut_head_tail_spaces (String.sub st 0 i))
      )
# 4902 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1306 "tools/coqdoc/cpretty.mll"
      ( None )
# 4907 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state

;;

# 1309 "tools/coqdoc/cpretty.mll"
 
  let coq_file f m =
    reset ();
    let c = open_in f in
    let lb = from_channel c in
      (Index.current_library := m;
       Output.initialize ();
       Output.start_module ();
       Output.start_coq (); coq_bol lb; Output.end_coq ();
       close_in c)

  let detect_subtitle f m =
    let c = open_in f in
    let lb = from_channel c in
    let sub = st_start m lb in
      close_in c;
      sub

# 4933 "tools/coqdoc/cpretty.ml"
